from __future__ import annotations

from typing import List, Optional, Set

import pytest
from pydantic import BaseModel, Field, ValidationError
from typing_extensions import Annotated

from metador_core.schema.partial import PartialFactory, PartialModel

from ..util.test_models import MyBaseModel

# from hypothesis import given, strategies as st


class ModelG(MyBaseModel):
    xyz: Optional[int]


class ModelD(MyBaseModel):
    class Config:
        frozen = True

    val: str
    val2: Optional[int]


class ModelE(ModelG):
    class Config:
        frozen = True

    prv: Annotated[ModelD, Field(description="simple model + annotation")]
    rec: Optional[ModelE]  # direct recursion
    col: Optional[Set[ModelD]]  # set
    fwd: List[ModelF] = []  # forward reference (+ indirect recursion)


class ModelF(ModelE):
    ...


class MyPartialFactory(PartialFactory):
    base_model = MyBaseModel


def test_partial_inheritance():
    # only for models sharing the selected base:
    class SomeModel(BaseModel):
        ...

    with pytest.raises(TypeError):
        MyPartialFactory.get_partial(SomeModel)

    classes = [ModelD, ModelE, ModelF, ModelG]
    for cls1 in classes:
        p1 = MyPartialFactory.get_partial(cls1)
        # partial is subclass of partial mixin and the base model:
        assert issubclass(p1, MyPartialFactory.base_model)
        assert issubclass(p1, MyPartialFactory.partial_mixin)
        #  partial MUST NOT be a subclass of original:
        assert not issubclass(p1, cls1)
        for cls2 in classes:
            p2 = MyPartialFactory.get_partial(cls2)
            if cls1 is cls2:
                # idempotent partial generation
                assert p1 is p2

            # but partials reflect inheritance of non-partials:
            orig_subclass = issubclass(cls1, cls2)
            partial_subclass = issubclass(p1, p2)
            assert orig_subclass == partial_subclass


# TODO: test partials properly


def test_partial_behavior():
    PartF = MyPartialFactory.get_partial(ModelF)

    # empty object is always possible
    empty_obj = PartF()

    # example object
    hlp = ModelF(prv=ModelD(val="c"))
    obj = ModelF(prv=ModelD(val="a"), fwd=[hlp.copy()], col=set([hlp.prv.copy()]))
    objE = ModelE(**obj.dict())

    # to_partial
    p_obj = PartF.to_partial(obj)  # no validation case
    assert PartF.to_partial(p_obj) is not p_obj  # should be fresh
    assert PartF.to_partial(objE) == p_obj
    assert PartF.to_partial(objE.dict()) == p_obj

    # test behavior with with invalid fields
    objInv = obj.copy(update=dict(prv=5))
    with pytest.raises(ValidationError):
        PartF.to_partial(objInv.dict())  # failed validation
    # with flag to drop invalid, it should work
    pObjInv = PartF.to_partial(objInv.dict(), ignore_invalid=True)
    # check that the result agrees with dropping the field (setting to "None")
    assert pObjInv.dict() == p_obj.copy(update={"prv": None}).dict()

    # cast
    assert PartF.cast(p_obj) is p_obj  # nothing to do -> returns original
    # should convert
    assert PartF.cast(obj) == p_obj
    p_objE = PartF.cast(objE)
    assert p_objE == p_obj
    assert p_objE is not p_obj

    # to partial -> from partial == original
    assert PartF.from_partial(p_obj) == obj

    # empty object is neutral element
    assert empty_obj.merge_with(p_obj, allow_overwrite=True) == p_obj
    assert p_obj.merge_with(empty_obj, allow_overwrite=True) == p_obj
    assert PartialModel.merge_with(empty_obj, empty_obj) == empty_obj
    assert PartialModel.merge(empty_obj, p_obj, empty_obj) == p_obj

    # without overwrite flag: cannot overwrite "a" with "a"
    with pytest.raises(ValueError):
        assert p_obj.merge_with(p_obj) == p_obj

    # merge 0 -> empty partial
    assert PartF.merge() == empty_obj

    # merge 1 -> partial with same content as given, implicit cast
    assert isinstance(PartF.merge(obj), PartF)
    assert PartF.merge(obj) == obj

    # merge 2 -> non-trivial merge
    hlp2 = ModelF(prv=ModelD(val="d"))
    obj2 = ModelF(
        prv=ModelD(val="b", val2=123),
        fwd=[hlp.copy(), hlp2.copy()],
        col=set([hlp.prv.copy(), hlp2.prv.copy()]),
    )
    obj3 = obj2.copy(update={"prv": obj.prv})
    merged = PartF.merge(obj2, obj3, allow_overwrite=True).from_partial()
    assert merged.col == obj2.col

    assert merged.prv.val == obj.prv.val  # overwritten value
    assert merged.prv.val2 == 123  # preserved by rec. merge
    # the list should be self-concatenated with deep merge
    assert merged.fwd == obj2.fwd + obj3.fwd



from __future__ import annotations

import sys
from typing import List, Literal, Optional, Set, Tuple, Union

import pytest
from pydantic import parse_obj_as
from typing_extensions import Annotated, ClassVar

from metador_core.util import typing as t

# test (enhanced) wrappers for normal checks


@pytest.mark.skipif(sys.version_info < (3, 9), reason="test needs Python >= 3.9")
def test_compat_get_origin():
    assert t.get_origin(list[int]) is t.List
    assert t.get_origin(set[int]) is t.Set
    assert t.get_origin(dict[int, str]) is t.Dict
    assert t.get_origin(type[int]) is t.Type


def test_hint_predicates():
    assert t.is_list(t.List[int])
    assert not t.is_list(t.Set[int])

    assert t.is_set(t.Set[int])
    assert not t.is_set(t.List[int])

    assert t.is_classvar(ClassVar[int])
    assert not t.is_classvar(int)

    assert t.is_nonetype(type(None))
    assert not t.is_nonetype(None)

    assert t.is_union(t.Union[int, str])
    assert t.is_union(Optional[int])  # Optional is a union
    assert t.is_union(t.Union[None, int])  # same as Optional
    assert not t.is_union(t.Union[int, int])

    assert t.is_optional(Optional[int])
    assert t.is_optional(Optional[t.Union[int, str]])
    assert t.is_optional(t.Union[None, int, str])  # union with None is Optional
    assert not t.is_optional(t.Union[int, str])

    assert t.is_annotated(Annotated[Optional[bool], "foo"])
    assert not t.is_annotated(Optional[Annotated[bool, "foo"]])
    assert not t.is_annotated(int)


class DummyClass:
    foo: Optional[str]


class DummyClass2(DummyClass):
    bar: int


def test_get_type_hints():
    # evaluated type hints
    th = t.get_type_hints(DummyClass2)
    assert th.get("foo") == Optional[str]
    assert th.get("bar") == int


def test_get_annotations():
    # unevaluated type hint
    ann = t.get_annotations(DummyClass2)
    assert "foo" not in ann
    assert ann.get("bar") == "int"

    # including inherited hints
    ann2 = t.get_annotations(DummyClass2, all=True)
    assert ann2.get("foo") == "Optional[str]"
    assert ann2.get("bar") == "int"


# test more specific stuff


def test_make_typehint():
    assert t.make_typehint(Literal[1], 5 == Literal[5])
    assert t.make_typehint(t.List[int], bool) == t.List[bool]
    assert t.make_typehint(t.Union[int, str], float, bool) == t.Union[float, bool]
    assert t.make_typehint(Optional[int], str) == t.Union[type(None), str]
    assert (
        t.make_typehint(Annotated[t.List[int], "hello"], bool)
        == Annotated[bool, "hello"]
    )


def test_make_literal():
    assert t.make_literal(None) == type(None)  # noqa: E721
    assert t.make_literal(True) == Literal[True]
    assert t.make_literal(1) == Literal[1]
    assert t.make_literal("foo") == Literal["foo"]

    # lists and tuples are mapped to tuples
    assert t.make_literal(["a", 1]) == t.Tuple[Literal["a"], Literal[1]]
    assert t.make_literal(("a", 1)) == t.Tuple[Literal["a"], Literal[1]]

    # test typeddict literal with pydantic
    TD = t.make_literal({"a": "bar"})
    assert parse_obj_as(TD, {"a": "bar"})
    with pytest.raises(ValueError):
        assert parse_obj_as(TD, {"a": "baz"})

    # sets and floats not supported
    with pytest.raises(ValueError):
        t.make_literal({1, 2, 3})
    with pytest.raises(ValueError):
        t.make_literal(3.12)


def test_traverse_typehint():
    u = Set[str]
    v, w = Tuple[int, bool], Annotated[u, "a"]
    x = Union[v, w]  # won't be visited, will be merged with y by type constructor
    y = Optional[x]
    z = List[y]
    hint = Annotated[z, "b"]

    expected_pre = [hint, z, y, v, int, bool, w, u, str, "a", type(None), "b"]
    assert list(t.traverse_typehint(hint)) == expected_pre

    expected_post = [int, bool, v, str, u, "a", w, type(None), y, z, "b", hint]
    assert list(t.traverse_typehint(hint, post_order=True)) == expected_post


def test_map_typehint():
    # map over structure of type hint, should preserve annotations
    orig = Tuple[int, List[Annotated[Union[int, Set[str]], "x"]]]
    exp = Tuple[bool, List[Annotated[Union[bool, Set[str]], "x"]]]

    def int_to_bool(x):
        return x if x is not int else bool

    assert t.map_typehint(orig, int_to_bool) == exp


def test_unoptional():
    # non-Optionals are just returned
    assert t.unoptional(int) == int
    assert t.unoptional(object) == object
    assert t.unoptional(type(None)) == type(None)  # noqa: E721

    # optional is unwrapped, nested type is untouched
    assert t.unoptional(Optional[int]) == int
    assert t.unoptional(Optional[t.List[str]]) == t.List[str]
    assert t.unoptional(Optional[t.Union[str, int]]) == t.Union[int, str]
    assert t.unoptional(Optional[Annotated[bool, "foo"]]) == Annotated[bool, "foo"]

    # annotation is preserved and transparent
    assert t.unoptional(Annotated[Optional[bool], "foo"]) == Annotated[bool, "foo"]


def test_is_subtype():
    # following is broken, should be fixed in typing_utils > 0.1.0
    # https://github.com/bojiang/typing_utils/pull/12
    # assert t.is_subtype(Literal["a"], Literal["a", "b"])

    assert t.is_subtype(int, t.Union[int, str])
    assert t.is_subtype(t.List[str], t.List[t.Union[t.Set[int], str]])
    assert t.is_subtype(Optional[int], t.Union[None, int, str])
    assert t.is_subtype(DummyClass2, DummyClass)
    assert t.is_subtype(List[int], Optional[List[Union[int, Set[str]]]])

    assert not t.is_subtype(DummyClass, DummyClass2)
    assert not t.is_subtype(Optional[int], t.Union[int, str])

    # Annotated types must agree on the type and annotation status, annotation is ignored
    # (at least currently this is the semantics)
    assert t.is_subtype(Annotated[int, "x"], Annotated[t.Union[int, str], "y"])
    assert not t.is_subtype(int, Annotated[t.Union[int, str], "y"])

    # Literal (custom hack)
    assert t.is_subtype(Literal[1], Literal[1, 2])
    assert t.is_subtype(Literal["b", "a"], Literal["a", "b"])
    assert t.is_subtype(Literal["a"], Literal["a", "b"])
    assert t.is_subtype(Literal["b"], Literal["a", "b"])
    assert not t.is_subtype(Literal["c"], Literal["a", "b"])
    assert not t.is_subtype(Literal["b", "a"], Literal["a"])


def test_predicates():
    d1, d2 = DummyClass(), DummyClass2()
    lst = [DummyClass, DummyClass2, d1, d2]

    class DummyClass3:
        ...

    lst2 = [DummyClass, DummyClass2, DummyClass3]

    # test predicates in a filter
    assert list(filter(t.is_subclass_of(DummyClass2), lst)) == [DummyClass2]
    assert list(filter(t.is_instance_of(DummyClass2), lst)) == [d2]
    assert list(filter(t.is_subtype_of(DummyClass), lst2)) == [DummyClass, DummyClass2]



import pytest

from metador_core.util import drop, pythonize_name
from metador_core.util.pytest import parameters, random_hex


@pytest.mark.parametrize("n", list(range(10)))
def test_random_hex(n):
    hx = random_hex(n)
    assert len(hx) == n


@pytest.mark.parametrize("n", list(range(5)))
def test_drop(n):
    lst = list(range(0, 3))
    assert list(drop(n, lst)) == lst[n:]


def test_parameters():
    assert parameters(1) == [(1,)]
    assert parameters([1, 2, 3]) == [(1,), (2,), (3,)]

    exp = [
        ("x", "y", "a"),
        ("x", "y", "b"),
        ("x", "y", "c", "d"),
        ("x", "y", "c", "e"),
        ("z", "f"),
    ]
    assert parameters({"x": {"y": ["a", "b", {"c": ["d", "e"]}]}, "z": "f"}) == exp


def test_pythonize_name():
    assert (
        pythonize_name("/Something, _-something + else")
        == "_Something____something___else"
    )
    assert pythonize_name("1no leading numbers") == "_1no_leading_numbers"
    assert pythonize_name("HE11o!") == "HE11o_"



"""Core Metadata schemas for Metador that are essential for the container API."""


from __future__ import annotations

import contextlib
from collections import ChainMap
from functools import partial
from typing import Any, ClassVar, Dict, Optional, Set, Type, cast

import wrapt
from pydantic import Extra, root_validator

from ..plugin.metaclass import PluginMetaclassMixin, UndefVersion
from ..util import cache, is_public_name
from ..util.models import field_atomic_types, traverse_typehint
from ..util.typing import (
    get_annotations,
    get_type_hints,
    is_classvar,
    is_instance_of,
    is_subclass_of,
    is_subtype,
)
from .base import BaseModelPlus
from .encoder import DynEncoderModelMetaclass
from .inspect import (
    FieldInspector,
    LiftedRODict,
    WrappedLiftedDict,
    lift_dict,
    make_field_inspector,
)

# from .jsonschema import finalize_schema_extra, schema_of
from .partial import PartialFactory, is_mergeable_type
from .types import to_semver_str


def add_missing_field_descriptions(schema, model):
    """Add missing field descriptions from own Fields info, if possible."""
    for fname, fjsdef in schema.get("properties", {}).items():
        if not fjsdef.get("description"):
            with contextlib.suppress(KeyError):
                if desc := model.Fields[fname].description:
                    fjsdef["description"] = desc


KEY_SCHEMA_PG = "$metador_plugin"
"""Key in JSON schema to put metador plugin name an version."""

KEY_SCHEMA_CONSTFLDS = "$metador_constants"
"""Key in JSON schema to put metador 'constant fields'."""


class SchemaBase(BaseModelPlus):
    __constants__: ClassVar[Dict[str, Any]]
    """Constant model fields, usually added with a decorator, ignored on input."""

    __overrides__: ClassVar[Set[str]]
    """Field names explicitly overriding inherited field type.

    Those not listed (by @overrides decorator) must, if they are overridden,
    be strict subtypes of the inherited type."""

    __types_checked__: ClassVar[bool]
    """Helper flag used by check_overrides to avoid re-checking."""

    class Config:
        @staticmethod
        def schema_extra(schema: Dict[str, Any], model: Type[BaseModelPlus]) -> None:
            model = UndefVersion._unwrap(model) or model

            # custom extra key to connect back to metador schema:
            if pgi := model.__dict__.get("Plugin"):
                schema[KEY_SCHEMA_PG] = pgi.ref().copy(exclude={"group"}).json_dict()

            # enrich schema with descriptions retrieved from e.g. docstrings
            if model is not MetadataSchema:
                add_missing_field_descriptions(schema, model)

            # special handling for "constant fields"
            if model.__constants__:
                schema[KEY_SCHEMA_CONSTFLDS] = {}
                for cname, cval in model.__constants__.items():
                    # list them (so they are not rejected even with additionalProperties=False)
                    schema["properties"][cname] = True
                    # store the constant alongside the schema
                    schema[KEY_SCHEMA_CONSTFLDS][cname] = cval

            # do magic (TODO: fix/rewrite)
            # finalize_schema_extra(schema, model, base_model=MetadataSchema)

    # NOTE: custom JSON schema feature is broken
    # @classmethod
    # def schema(cls, *args, **kwargs):
    #    """Return customized JSONSchema for this model."""
    #    return schema_of(UndefVersion._unwrap(cls) or cls, *args, **kwargs)

    @root_validator(pre=True)
    def override_consts(cls, values):
        """Override/add defined schema constants.

        They must be present on dump, but are ignored on load.
        """
        values.update(cls.__constants__)
        return values


ALLOWED_SCHEMA_CONFIG_FIELDS = {"title", "extra", "allow_mutation"}
"""Allowed pydantic Config fields to be overridden in schema models."""


class SchemaMagic(DynEncoderModelMetaclass):
    """Metaclass for doing some magic."""

    def __new__(cls, name, bases, dct):
        # enforce single inheritance
        if len(bases) > 1:
            raise TypeError("A schema can only have one parent schema!")
        baseschema = bases[0]

        # only allow inheriting from other schemas:
        # NOTE: can't normally happen (this metaclass won't be triggered)
        # if not issubclass(baseschema, SchemaBase):
        #     raise TypeError(f"Base class {baseschema} is not a MetadataSchema!")

        # prevent user from defining special schema fields by hand
        for atr in SchemaBase.__annotations__.keys():
            if atr in dct:
                raise TypeError(f"{name}: Invalid attribute '{atr}'")

        # prevent most changes to pydantic config
        if conf := dct.get("Config"):
            for conffield in conf.__dict__:
                if (
                    is_public_name(conffield)
                    and conffield not in ALLOWED_SCHEMA_CONFIG_FIELDS
                ):
                    raise TypeError(f"{name}: {conffield} must not be set or changed!")

        # generate pydantic model of schema (further checks are easier that way)
        # can't do these checks in __init__, because in __init__ the bases could be mangled
        ret = super().__new__(cls, name, bases, dct)

        # prevent user defining fields that are constants in a parent
        if base_consts := set(getattr(baseschema, "__constants__", {}).keys()):
            new_defs = set(get_annotations(ret).keys())
            if illegal := new_defs.intersection(base_consts):
                msg = (
                    f"{name}: Cannot define {illegal}, defined as const field already!"
                )
                raise TypeError(msg)

        # prevent parent-compat breaking change of extra handling / new fields:
        parent_forbids_extras = baseschema.__config__.extra is Extra.forbid
        if parent_forbids_extras:
            # if parent forbids, child does not -> problem (child can parse, parent can't)
            extra = ret.__config__.extra
            if extra is not Extra.forbid:
                msg = (
                    f"{name}: cannot {extra.value} extra fields if parent forbids them!"
                )
                raise TypeError(msg)

            # parent forbids extras, child has new fields -> same problem
            if new_flds := set(ret.__fields__.keys()) - set(
                baseschema.__fields__.keys()
            ):
                msg = f"{name}: Cannot define new fields {new_flds} if parent forbids extra fields!"
                raise TypeError(msg)

        # everything looks ok
        return ret

    def __init__(self, name, bases, dct):
        self.__types_checked__ = False  # marker used by check_types (for performance)

        # prevent implicit inheritance of class-specific internal/meta stuff:
        # should be taken care of by plugin metaclass
        assert self.Plugin is None or self.Plugin != bases[0].Plugin

        # also prevent inheriting override marker
        self.__overrides__ = set()

        # and manually prevent inheriting annotations (for Python < 3.10)
        if "__annotations__" not in self.__dict__:
            self.__annotations__ = {}

        # "constant fields" are inherited, but copied - not shared
        self.__constants__ = {}
        for b in bases:
            self.__constants__.update(getattr(b, "__constants__", {}))

    @property  # type: ignore
    @cache  # noqa: B019
    def _typehints(self):
        """Return typehints of this class."""
        return get_type_hints(self)

    @property  # type: ignore
    @cache  # noqa: B019
    def _base_typehints(self):
        """Return typehints accumulated from base class chain."""
        return ChainMap(
            *(b._typehints for b in self.__bases__ if issubclass(b, SchemaBase))
        )

    # ---- for public use ----

    def __str__(self):
        """Show schema and field documentation."""
        unwrapped = UndefVersion._unwrap(self)
        schema = unwrapped or self
        defstr = f"Schema {super().__str__()}"
        defstr = f"{defstr}\n{'='*len(defstr)}"
        descstr = ""
        if schema.__doc__ is not None and schema.__doc__.strip():
            desc = schema.__doc__
            descstr = f"\nDescription:\n------------\n\t{desc}"
        fieldsstr = f"Fields:\n-------\n\n{str(self.Fields)}"
        return "\n".join([defstr, descstr, fieldsstr])

    @property
    def Fields(self: Any) -> Any:
        """Access the field introspection interface."""
        fields = make_schema_inspector(self)
        # make sure that subschemas accessed in a schema without explicit version
        # are also marked so we can check if someone uses them illegally
        if issubclass(self, UndefVersion):
            return WrappedLiftedDict(fields, UndefVersionFieldInspector)
        else:
            return fields

    @property
    def Partial(self):
        """Access the partial schema based on the current schema."""
        return PartialSchemas.get_partial(self)


class SchemaMetaclass(PluginMetaclassMixin, SchemaMagic):
    """Combine schema magic with general plugin magic."""


class MetadataSchema(SchemaBase, metaclass=SchemaMetaclass):
    """Extends Pydantic models with custom serializers and functions."""

    Plugin: ClassVar[Optional[Type]]  # user-defined inner class (for schema plugins)


# ----


def _indent_text(txt, prefix="\t"):
    """Add indentation at new lines in given string."""
    return "\n".join(map(lambda line: f"{prefix}{line}", txt.split("\n")))


class SchemaFieldInspector(FieldInspector):
    """MetadataSchema-specific field inspector.

    It adds a user-friendly repr and access to nested subschemas.
    """

    schemas: LiftedRODict
    _origin_name: str

    def _get_description(self):
        fld = self.origin._typehints[self.name]
        if any(map(lambda x: fld is x, (None, type(None), bool, int, float, str))):
            return None
        return super()._get_description()

    def __init__(self, model: Type[MetadataSchema], name: str, hint: str):
        super().__init__(model, name, hint)

        # to show plugin name and version in case of registered plugin schemas:
        og = self.origin
        self._origin_name = f"{og.__module__}.{og.__qualname__}"
        if pgi := og.Plugin:
            self._origin_name += f" (plugin: {pgi.name} {to_semver_str(pgi.version)})"

        # access to sub-entities/schemas:
        subschemas = list(field_atomic_types(og.__fields__[name], bound=MetadataSchema))
        self.schemas = lift_dict("Schemas", {s.__name__: s for s in set(subschemas)})

    def __repr__(self) -> str:
        desc_str = ""
        if self.description:
            desc_str = f"description:\n{_indent_text(self.description)}\n"
        schemas_str = (
            f"schemas: {', '.join(iter(self.schemas))}\n" if self.schemas else ""
        )
        info = f"type: {str(self.type)}\norigin: {self._origin_name}\n{schemas_str}{desc_str}"
        return f"{self.name}\n{_indent_text(info)}"


def _is_schema_field(schema, key: str):
    """Return whether a given key name is a non-constant model field."""
    return key in schema.__fields__ and key not in schema.__constants__


@cache
def make_schema_inspector(schema):
    return make_field_inspector(
        schema,
        "Fields",
        bound=MetadataSchema,
        key_filter=partial(_is_schema_field, schema),
        i_cls=SchemaFieldInspector,
    )


# ----
# wrapper to "infect" nested schemas with UndefVersion


class UndefVersionFieldInspector(wrapt.ObjectProxy):
    @property
    def schemas(self):
        return WrappedLiftedDict(self.__wrapped__.schemas, UndefVersion._mark_class)

    def __repr__(self):
        return repr(self.__wrapped__)


# ----


class PartialSchemas(PartialFactory):
    """Partial model for MetadataSchema model.

    Needed for harvesters to work (which can provide validated but partial metadata).
    """

    base_model = SchemaBase

    # override to ignore "constant fields" for partials
    @classmethod
    def _get_field_vals(cls, obj):
        return (
            (k, v)
            for k, v in super()._get_field_vals(obj)
            if k not in obj.__constants__
        )

    # override to add some fixes to partials
    @classmethod
    def _create_partial(cls, mcls, *, typehints=...):
        th = getattr(mcls, "_typehints", None)
        unw = UndefVersion._unwrap(mcls) or mcls
        ret, nested = super()._create_partial(unw, typehints=th)
        # attach constant field list for field filtering
        ret.__constants__ = getattr(mcls, "__constants__", set())
        # copy custom parser to partial
        # (otherwise partial can't parse correctly with parser mixin)
        if parser := getattr(mcls, "Parser", None):
            ret.Parser = parser
        return (ret, nested)


# --- delayed checks (triggered during schema loading) ---
# if we do it earlier, might lead to problems with forward refs and circularity


def check_types(schema: Type[MetadataSchema], *, recheck: bool = False):
    if schema is MetadataSchema or schema.__types_checked__ and not recheck:
        return
    schema.__types_checked__ = True

    # recursively check compositional and inheritance dependencies
    for b in schema.__bases__:
        if issubclass(b, MetadataSchema):
            check_types(b, recheck=recheck)

    schemaFields = cast(Any, schema.Fields)
    for f in schemaFields:  # type: ignore
        for sname in schemaFields[f].schemas:
            s = schemaFields[f].schemas[sname]
            if s is not schema and issubclass(s, MetadataSchema):
                check_types(s, recheck=recheck)

    check_allowed_types(schema)
    check_overrides(schema)


def check_allowed_types(schema: Type[MetadataSchema]):
    """Check that shape of defined fields is suitable for deep merging."""
    hints = cast(Any, schema._typehints)
    for field, hint in hints.items():
        if not is_public_name(field):
            continue  # private field
        if not is_mergeable_type(hint):
            msg = f"{schema}:\n\ttype of '{field}' contains a forbidden pattern!"
            raise TypeError(msg)

        # check that no nested schemas from undefVersion plugins are used in field definitions
        # (the Plugin metaclass cannot check this, but it checks for inheritance)
        if illegal := next(
            filter(
                is_subclass_of(UndefVersion),
                filter(is_instance_of(type), traverse_typehint(hint)),
            ),
            None,
        ):
            msg = f"{schema}:\n\ttype of '{field}' contains an illegal subschema:\n\t\t{illegal}"
            raise TypeError(msg)


def infer_parent(plugin: Type[MetadataSchema]) -> Optional[Type[MetadataSchema]]:
    """Return closest base schema that is a plugin, or None.

    This allows to skip over intermediate schemas and bases that are not plugins.
    """
    return next(
        filter(
            lambda c: issubclass(c, MetadataSchema) and c.__dict__.get("Plugin"),
            plugin.__mro__[1:],
        ),
        None,
    )


def is_pub_instance_field(schema, name, hint):
    """Return whether field `name` in `schema` is a non-constant, public schema instance field."""
    return (
        is_public_name(name)
        and not is_classvar(hint)
        and name not in schema.__constants__
    )


def detect_field_overrides(schema: Type[MetadataSchema]):
    anns = get_annotations(schema)
    base_hints = cast(Any, schema._base_typehints)
    new_hints = {n for n, h in anns.items() if is_pub_instance_field(schema, n, h)}
    return set(base_hints.keys()).intersection(new_hints)


def check_overrides(schema: Type[MetadataSchema]):
    """Check that fields are overridden to subtypes or explicitly declared as overridden."""
    hints = cast(Any, schema._typehints)
    base_hints = cast(Any, schema._base_typehints)

    actual_overrides = detect_field_overrides(schema)
    undecl_override = actual_overrides - schema.__overrides__
    if unreal_override := schema.__overrides__ - set(base_hints.keys()):
        msg = f"{schema.__name__}: No parent field to override: {unreal_override}"
        raise ValueError(msg)

    if miss_override := schema.__overrides__ - actual_overrides:
        msg = f"{schema.__name__}: Missing claimed field overrides: {miss_override}"
        raise ValueError(msg)

    # all undeclared overrides must be strict subtypes of the inherited type:
    for fname in undecl_override:
        hint, parent_hint = hints[fname], base_hints[fname]
        if not is_subtype(hint, parent_hint):
            parent = infer_parent(schema)
            parent_name = (
                parent.Fields[fname]._origin_name
                if parent
                else schema.__base__.__name__
            )
            msg = f"""The type assigned to field '{fname}'
in schema {repr(schema)}:

  {hint}

does not look like a valid subtype of the inherited type:

  {parent_hint}

from schema {parent_name}.

If you are ABSOLUTELY sure that this is a false alarm,
use the @overrides decorator to silence this error
and live forever with the burden of responsibility.
"""
            raise TypeError(msg)


# def detect_field_overrides(schema: Type[MetadataSchema]) -> Set[str]:
#     return {n
#         for n in updated_fields(schema)
#         if is_public_name(n) and not is_classvar(schema._typehints[n]) and n not in schema.__constants__
#     }



from typing import Union

import h5py
import numpy as np
import pytest

from metador_core.ih5.container import (
    IH5AttributeManager,
    IH5Dataset,
    IH5Group,
    IH5Record,
)
from metador_core.ih5.overlay import DEL_VALUE, SUBST_KEY, H5Type, IH5Node
from metador_core.ih5.skeleton import IH5Skeleton, SkeletonNodeInfo


def create_entries(node: Union[IH5Group, IH5AttributeManager]):
    node["array"] = [0, 0, 0]
    node["bool"] = False
    node["int"] = 0
    node["string"] = node._gpath + ("@" if node.__is_attrs__ else "/") + "string"
    node["raw"] = np.void(b"raw")


def fill_dummy_ds(ds: IH5Record, flat: bool):
    def next_patch():
        if not flat:
            ds.commit_patch()
            ds.create_patch()

    create_entries(ds["/"])
    create_entries(ds.attrs)
    create_entries(ds["int"].attrs)
    next_patch()
    for name in ["a", "b"]:
        grp = ds.create_group(name)
        create_entries(grp)
        create_entries(grp.attrs)
        create_entries(grp["array"].attrs)
        next_patch()
        for subname in ["a", "b"]:
            sgrp = grp.create_group(subname)
            create_entries(sgrp)
            create_entries(sgrp["array"].attrs)
            next_patch()


# NOTE: this might be slowing down this whole module, maybe consider caching
# or keeping a bunch of pre-generated test-files to be used with the testinput fixture


@pytest.fixture
def dummy_ds_factory(tmp_ds_path_factory):
    """Return record with base container still in writable mode."""
    records = []

    def get_ds(flat: bool, commit: bool):
        ds = IH5Record(tmp_ds_path_factory(), "w")
        fill_dummy_ds(ds, flat)
        if commit:
            ds.commit_patch()
        records.append(ds)
        return ds

    yield get_ds
    for ds in records:
        ds.close()


# --------


def test_ds_ndim(tmp_ds_path):
    """Make sure ndim is passed through correctly."""
    with IH5Record(tmp_ds_path, "w") as ds:
        create_entries(ds["/"])
        ds["mat"] = [[0, 1, 2], [3, 4, 5]]

        assert ds["raw"].ndim == 0
        assert ds["bool"].ndim == 0
        assert ds["array"].ndim == 1
        assert ds["mat"].ndim == 2


def test_node_instance_checks(tmp_ds_path):
    # test that instance creation expectedly fails due to sanity checks
    with IH5Record(tmp_ds_path, "w") as ds:
        with pytest.raises(ValueError):
            IH5Node(ds, "", 0)  # type: ignore
        with pytest.raises(ValueError):
            IH5Node(ds, "/bla", -1)  # type: ignore
        with pytest.raises(ValueError):
            IH5Node(ds, "bla", 1)  # type: ignore
        IH5Node(ds, "/bla", 0)  # type: ignore


def test_ih5node_compare(tmp_ds_path_factory):
    # nodes from same open record with same path and creation index should be equal
    # everything else should not.
    ds1name, ds2name = tmp_ds_path_factory(), tmp_ds_path_factory()
    with IH5Record(ds1name, "w") as ds1:
        ds1["/foo"] = "bar"
        foo = ds1["/foo"]

        assert ds1["/"] == ds1["/"]
        assert ds1["/foo"] == foo
        assert ds1["/"] != foo  # different path

        ds1.commit_patch()
        ds1.create_patch()
        assert ds1["/foo"] == foo  # still same patch
        del ds1["/foo"]
        ds1["/foo"] = "blub"
        assert ds1["/foo"] != foo  # now different patch

        with IH5Record(ds2name, "w") as ds2:
            ds2["/foo"] = "bar"
            assert ds2["/"] == ds2["/"]
            assert ds2["/foo"] != foo  # different record


def test_abs_rel_parent_path(tmp_ds_path):
    with IH5Record(tmp_ds_path, "w") as ds:
        ds["grp/foo/bar"] = 123

        assert ds["grp"]._abs_path("foo") == "/grp/foo"
        assert ds["grp"]._abs_path("/foo") == "/foo"
        assert ds["/"]._abs_path("foo") == "/foo"
        assert ds["/"]._abs_path("/") == "/"

        assert ds["grp"]._rel_path("/grp/foo/bar") == "foo/bar"
        assert ds["grp"]._rel_path("foo") == "foo"
        assert ds["/"]._rel_path("/foo/bar") == "foo/bar"
        assert ds["/"]._rel_path("/foo/bar") == "foo/bar"
        with pytest.raises(RuntimeError):
            ds["grp"]._rel_path("/invalid")

        assert ds["grp/foo/bar"]._parent_path() == "/grp/foo"  # type: ignore
        assert ds["grp/foo"]._parent_path() == "/grp"
        assert ds["/"]._parent_path() == "/"


def test_check_key(tmp_ds_path):
    # attributes cannot have slashes, paths cannot be empty
    with IH5Record(tmp_ds_path, "w") as ds:
        with pytest.raises(ValueError):
            ds[""] = 1
        with pytest.raises(ValueError):
            ds["invalid@path"] = 1
        with pytest.raises(ValueError):
            ds["hello world"] = 1
        with pytest.raises(ValueError):
            ds["абвгд"] = 1
        with pytest.raises(ValueError):
            ds.attrs["invalid@attr"] = 1
        with pytest.raises(ValueError):
            ds.attrs[""] = 1
        with pytest.raises(ValueError):
            ds.attrs["/invalidattr"] = 1
        with pytest.raises(ValueError):
            ds.attrs["invalid/attr"] = 1


def test_latest_container_idx(tmp_ds_path):
    # test finding most recent entity at a path
    with IH5Record(tmp_ds_path, "w") as ds:
        ds["foo/bar"] = 123
        ds.commit_patch()
        assert IH5Group._latest_idx(ds._files, "/foo/bar") == 0
        ds.create_patch()
        del ds["foo/bar"]
        ds.commit_patch()
        assert IH5Group._latest_idx(ds._files, "/foo/bar") is None
        ds.create_patch()
        ds["foo/bar"] = 456
        ds.commit_patch()
        assert IH5Group._latest_idx(ds._files, "/foo/bar") == 2


def test_visit(tmp_ds_path):
    with IH5Record(tmp_ds_path, "w") as ds:
        ds["grp/foo/bar"] = 123
        ds.attrs["rootattr"] = "yay"  # type: ignore
        ds["grp/foo/bar"].attrs["someattr"] = "value"  # type: ignore
        ds.create_group("grp/qux")

        lst = []
        # check expected return value
        ret = ds["/grp"].visit(lambda x: lst.append(x) if not lst else lst[0])
        assert ret == lst[0]

        lst.clear()
        # collect all inside group 'grp', return None -> no early return
        ds["/grp"].visit(lst.append)
        # should be relative paths in alphabetical order
        assert lst == ["foo", "foo/bar", "qux"]

        # same again, from root node
        lst.clear()
        ds.visit(lst.append)
        assert lst == ["grp", "grp/foo", "grp/foo/bar", "grp/qux"]
        assert IH5Skeleton.for_record(ds) == IH5Skeleton(
            __root__={
                "/": SkeletonNodeInfo(
                    node_type=H5Type.group, patch_index=0, attrs={"rootattr": 0}
                ),
                "/grp": SkeletonNodeInfo(
                    node_type=H5Type.group, patch_index=0, attrs={}
                ),
                "/grp/foo": SkeletonNodeInfo(
                    node_type=H5Type.group, patch_index=0, attrs={}
                ),
                "/grp/foo/bar": SkeletonNodeInfo(
                    node_type=H5Type.dataset, patch_index=0, attrs={"someattr": 0}
                ),
                "/grp/qux": SkeletonNodeInfo(
                    node_type=H5Type.group, patch_index=0, attrs={}
                ),
            }
        )


def test_set_inside_value(tmp_ds_path):
    with IH5Record(tmp_ds_path, "w") as ds:
        ds["a"] = [1, 2, 3]
        a = ds["a"]
        ds.commit_patch()

        # cannot set, no current patch
        with pytest.raises(ValueError) as e:
            a[2] = 5
        assert str(e).lower().find("patch") >= 0

        with pytest.raises(ValueError) as e:
            ds["a"].copy_into_patch()  # type: ignore
        assert str(e).lower().find("patch") >= 0

        ds.create_patch()
        # cannot set using this node (node from previous patch)
        with pytest.raises(ValueError) as e:
            a[2] = 5
        assert str(e).lower().find("latest") >= 0

        # still cannot write (no value there)
        with pytest.raises(ValueError) as e:
            ds["a"][2] = 5
        assert str(e).lower().find("latest") >= 0

        # explicitly copy value into new patch
        a.copy_into_patch()  # type: ignore
        with pytest.raises(ValueError) as e:
            ds["a"].copy_into_patch()  # type: ignore
        assert str(e).lower().find("latest") >= 0

        ds["a"][2] = 5  # now we can modify it
        assert np.array_equal(ds["a"][()], np.array([1, 2, 5]))  # type: ignore
        ds["a"][0:2] = [3, 4]
        assert np.array_equal(ds["a"][()], np.array([3, 4, 5]))  # type: ignore


def test_create_virtual_fail(tmp_ds_path):
    # _create_virtual should fail when the path already exists
    with IH5Record(tmp_ds_path, "w") as ds:
        ds["a/b/c"] = 123
        assert not ds["a"]._create_virtual("b")  # should not work, is a group
        assert not ds["a"]._create_virtual("b/c")  # should not work, is a value

        ds.commit_patch()
        ds.create_patch()
        del ds["a/b/c"]
        assert "a/b/c" not in ds
        assert ds["a"]._create_virtual("b/c")  # should work, ignoring deletion mark


def test_forbidden_entities_fail(tmp_ds_path):
    with IH5Record(tmp_ds_path, "w") as ds:
        # we should not be able to set a DEL value or a SUB attr key
        with pytest.raises(ValueError) as e:
            ds["a/b/c"] = DEL_VALUE
        assert str(e).lower().find("forbidden") >= 0
        with pytest.raises(ValueError) as e:
            ds.attrs[SUBST_KEY] = 123
        assert str(e).lower().find("invalid") >= 0

        # hard links to values and groups
        grp = ds.create_group("grp")
        grp["val"] = 123
        with pytest.raises(ValueError) as e:
            ds["a/b/c"] = grp
        assert str(e).lower().find("hard links") >= 0
        with pytest.raises(ValueError) as e:
            ds["a/b/c"] = grp["val"]
        assert str(e).lower().find("hard links") >= 0

        # symbolic and external links
        with pytest.raises(ValueError) as e:
            ds["a/b/c"] = h5py.SoftLink("/grp")
        assert str(e).lower().find("symlink") >= 0
        with pytest.raises(ValueError) as e:
            ds["a/b/c"] = h5py.ExternalLink("file.h5", "/grp")
        assert str(e).lower().find("externallink") >= 0


def test_modify_readonly_fail(dummy_ds_factory):
    # modifying record values or attributes should not work while not creating patch
    def expected_msg(e):
        return str(e).lower().find("patch") >= 0

    ds1 = dummy_ds_factory(flat=True, commit=True)
    ds2 = dummy_ds_factory(flat=False, commit=True)
    for ds in [ds1, ds2]:
        a = ds["a"]
        assert isinstance(a, IH5Group)

        # set a value
        with pytest.raises(ValueError) as e:
            a["newkey"] = "value"
        assert expected_msg(e)
        with pytest.raises(ValueError) as e:
            a.attrs["newkey"] = "value"
        assert expected_msg(e)

        # create a group
        with pytest.raises(ValueError) as e:
            ds.create_group("c")
        assert expected_msg(e)

        # try deleting
        with pytest.raises(ValueError) as e:
            del a["a"]
        with pytest.raises(ValueError) as e:
            del a["bool"]
        assert expected_msg(e)
        with pytest.raises(ValueError) as e:
            del a.attrs["newkey"]
        assert expected_msg(e)


def test_create_over_existing_fail(dummy_ds_factory):
    # overwriting records or groups should not work without deleting first (like h5py)
    def expected_msg(e):
        return str(e).lower().find("exist") >= 0

    ds1 = dummy_ds_factory(flat=True, commit=False)
    ds2 = dummy_ds_factory(flat=False, commit=False)
    for ds in [ds1, ds2]:
        with pytest.raises(ValueError) as e:
            ds["a"] = True  # value over group
        expected_msg(e)
        with pytest.raises(ValueError) as e:
            ds["/bool"] = True  # value over value
        expected_msg(e)
        with pytest.raises(ValueError) as e:
            ds.create_group("/a")  # group over group
        expected_msg(e)
        with pytest.raises(ValueError) as e:
            ds.create_group("bool")  # group over value
        expected_msg(e)


def test_delete_nonexisting_fail(dummy_ds_factory):
    # deleting non-existing entities should fail
    def expected_msg(e):
        return str(e).lower().find("exist") >= 0

    ds1 = dummy_ds_factory(flat=True, commit=False)
    ds2 = dummy_ds_factory(flat=False, commit=False)
    for ds in [ds1, ds2]:
        with pytest.raises(KeyError) as e:
            del ds["missing"]
        expected_msg(e)
        with pytest.raises(KeyError) as e:
            del ds["/missing"]
        expected_msg(e)

        # delete real thing, but twice
        del ds["a"]
        with pytest.raises(KeyError) as e:
            del ds["a"]
        del ds["bool"]
        with pytest.raises(KeyError) as e:
            del ds["bool"]
        del ds.attrs["bool"]
        with pytest.raises(KeyError) as e:
            del ds.attrs["bool"]


def test_create_access_relative_absolute(tmp_ds_path):
    # create datasets/groups using relative and absolute paths
    with IH5Record(tmp_ds_path, "w") as ds:
        nested = ds.create_group("nested")
        nested.create_group("deep")
        nested.create_group("/toplevel")
        nested["data"] = 123
        nested["data"].attrs["key"] = "value"  # type: ignore
        nested["/moredata"] = 456
        ds.commit_patch()

        assert IH5Skeleton.for_record(ds) == IH5Skeleton(
            __root__={
                "/": SkeletonNodeInfo(node_type=H5Type.group, patch_index=0, attrs={}),
                "/moredata": SkeletonNodeInfo(
                    node_type=H5Type.dataset, patch_index=0, attrs={}
                ),
                "/nested": SkeletonNodeInfo(
                    node_type=H5Type.group, patch_index=0, attrs={}
                ),
                "/nested/data": SkeletonNodeInfo(
                    node_type=H5Type.dataset, patch_index=0, attrs={"key": 0}
                ),
                "/nested/deep": SkeletonNodeInfo(
                    node_type=H5Type.group, patch_index=0, attrs={}
                ),
                "/toplevel": SkeletonNodeInfo(
                    node_type=H5Type.group, patch_index=0, attrs={}
                ),
            }
        )
        # access from inside using absolute or relative path
        assert ds["nested"]["data"]._gpath == "/nested/data"
        assert ds["nested"]["/moredata"]._gpath == "/moredata"
        # cannot access into dataset
        with pytest.raises(ValueError) as e:
            ds["nested/data/something"]
        assert str(e).lower().find("value") >= 0
        # cannot access, no such entity
        with pytest.raises(KeyError) as e:
            ds["nested/missing"]


def test_fresh_patch_overlay(dummy_ds_factory):
    # create a new patch with various changes to the dummy data
    ds = dummy_ds_factory(flat=False, commit=True)
    ds.create_patch()

    atrs = ds["a"].attrs
    del atrs["bool"]  # delete an attribute (will implicitly create virtual node)
    atrs["int"] = "surprise"  # overwrite an attribute
    atrs["key"] = True  # add an attribute

    del ds["b"]  # delete a group
    del ds["bool"]  # delete a value

    del ds["a/a"]  # overwrite a group with a value
    ds["a/a"] = 123

    del ds["a/b"]  # overwrite a group with a fresh group
    ds.create_group("a/b")

    del ds["a/bool"]  # overwrite a value with a new value
    ds["/a/bool"] = "new value"

    del ds["a/array"]  # overwrite a value with a fresh group with contents
    assert "a/array" not in ds
    ds["a/array/data"] = 456  # implicit group creation!
    assert "a/array" in ds
    assert "a/array/data" in ds

    ds["a"].create_group("c")  # create new group
    ds["a/c/d"] = 789  # create new value

    # set an attribute
    ds["a/c/d"].attrs["key"] = 1337  # type: ignore

    ds.commit_patch()

    # check expected values
    assert "b" not in ds
    assert "bool" not in ds
    assert "a/a" in ds and ds["a"]["a"][()] == 123
    assert isinstance(ds["a/b"], IH5Group) and list(ds["a/b"].attrs.keys()) == []
    assert ds["a/bool"][()] == b"new value"
    assert isinstance(ds["a/array"], IH5Group)
    assert isinstance(ds["a/array/data"], IH5Dataset)
    assert isinstance(ds["/a/c"], IH5Group)
    assert isinstance(ds["/a/c/d"], IH5Dataset) and ds["a"]["c/d"][()] == 789
    assert ds["a/c/d"].attrs["key"] == 1337  # type: ignore
    assert "bool" not in ds["a"].attrs
    assert ds["a"].attrs["int"] == "surprise"
    assert ds["a"].attrs["key"] == True


def test_clear_all_override(dummy_ds_factory):
    # get multi-patch record, delete EVERYTHING
    ds = dummy_ds_factory(flat=False, commit=True)

    def clear_and_verify():
        for k in ds.keys():
            del ds[k]
        for k in ds.attrs.keys():
            del ds.attrs[k]

        # nothing should be there
        assert list(ds.attrs.keys()) == []
        assert list(ds.keys()) == []

    def create_and_verify(pidx):
        ds.create_group("a")
        ds["b/a"] = 123
        ds["/b"]["a"].attrs["atr"] = "string"  # type: ignore
        ds.create_group("b/b")

        # only the new stuff should be there
        assert IH5Skeleton.for_record(ds) == IH5Skeleton(
            __root__={
                "/": SkeletonNodeInfo(node_type=H5Type.group, patch_index=0, attrs={}),
                "/a": SkeletonNodeInfo(
                    node_type=H5Type.group, patch_index=pidx, attrs={}
                ),
                "/b": SkeletonNodeInfo(
                    node_type=H5Type.group, patch_index=pidx, attrs={}
                ),
                "/b/a": SkeletonNodeInfo(
                    node_type=H5Type.dataset, patch_index=pidx, attrs={"atr": pidx}
                ),
                "/b/b": SkeletonNodeInfo(
                    node_type=H5Type.group, patch_index=pidx, attrs={}
                ),
            }
        )

    # fully clear and refill within the same patch
    ds.create_patch()
    clear_and_verify()
    create_and_verify(ds._ublock(-1).patch_index)
    ds.commit_patch()
    # clear in one patch, overwrite in another
    ds.create_patch()
    clear_and_verify()
    ds.commit_patch()
    ds.create_patch()
    create_and_verify(ds._ublock(-1).patch_index)


def test_create_value_patch_delete_value_delete_parent_group(tmp_ds_path):
    # create value inside a group, in next patch remove it,
    # then try removing the corresponding parent group
    # based on actual bug (group removal failed, but should succeed)
    with IH5Record(tmp_ds_path, "w") as ds:
        ds["a/b/c"] = 123
        ds.commit_patch()
        ds.create_patch()

        assert ds["a/b/c"][()] == 123  # type: ignore
        del ds["a/b/c"]
        assert "a/b/c" not in ds

        assert "a/b" in ds
        del ds["a/b"]
        assert "a/b" not in ds

        ds["a/b"] = 456
        assert ds["a/b"][()] == 456  # type: ignore
        assert "a/b" in ds


def test_not_open_fail(dummy_ds_factory):
    ds = dummy_ds_factory(flat=False, commit=False)
    a = ds["a"]
    ds.close()

    def assert_ex(f):
        with pytest.raises(ValueError) as e:
            f()
        assert str(e).lower().find("not open") >= 0

    # check that public methods fail gracefully when record not open
    assert_ex(lambda: a.__delitem__("b"))
    assert_ex(lambda: a.__setitem__("x/y", "z"))
    assert_ex(lambda: a.get("b"))
    assert_ex(lambda: a["b"])
    assert_ex(lambda: a.attrs)
    assert_ex(lambda: a.keys())
    assert_ex(lambda: a.values())
    assert_ex(lambda: a.items())
    assert_ex(lambda: a.create_group("some_group"))
    assert_ex(lambda: a.visit(print))
    assert_ex(lambda: "b" in a)
    assert_ex(lambda: iter(a))



import enum
from typing import Literal, Optional

import pytest
from pydantic import BaseModel, ValidationError

from metador_core.plugin.util import register_in_group
from metador_core.schema import MetadataSchema
from metador_core.schema.decorators import add_const_fields, make_mandatory, override


def test_schema_base_check():
    """Should refuse to work on non-metador models."""

    class NonSchema(BaseModel):
        foo: Optional[int]

    with pytest.raises(TypeError):
        make_mandatory("foo")(NonSchema)

    with pytest.raises(TypeError):
        add_const_fields({"bla": None})(NonSchema)

    with pytest.raises(TypeError):
        override("foo")(NonSchema)


@pytest.fixture
def dummy_schema():
    class DummySchema(MetadataSchema):
        foo: Optional[int]

    return DummySchema


def test_field_private_check(dummy_schema):
    """Should prevent messing with private fields."""
    with pytest.raises(ValueError):
        make_mandatory("foo", "_bar")(dummy_schema)

    with pytest.raises(ValueError):
        add_const_fields({"bla": int, "_bar": int})(dummy_schema)

    with pytest.raises(ValueError):
        override("_bar")(dummy_schema)


def test_add_const_fields(dummy_schema):
    assert dummy_schema(foo=5).dict().get("foo") == 5

    # new constant field works
    add_const_fields({"bar": "world"})(dummy_schema)
    assert dummy_schema(foo=5).dict().get("bar") == "world"

    with pytest.raises(ValueError):
        # override = False, foo exists
        add_const_fields({"foo": "blub"})(dummy_schema)

    # override into a constant field works
    add_const_fields({"foo": "hello"}, override=True)(dummy_schema)
    assert dummy_schema(foo=5).dict().get("foo") == "hello"

    # preserved in subclass
    class Child(dummy_schema):
        ...

    assert Child(foo=5).dict().get("foo") == "hello"

    # behaves like parent type in foo
    class ChildWorks(dummy_schema):
        qux: str

    assert ChildWorks(qux="x", foo=5).dict().get("foo") == "hello"
    assert ChildWorks(qux="y").dict().get("bar") == "world"

    # const can be overridden with other const
    add_const_fields({"foo": "bye"}, override=True)(ChildWorks)
    assert ChildWorks(qux="x", foo=5).dict().get("foo") == "bye"

    # cannot override constant with normal field again
    with pytest.raises(TypeError):

        class ChildFail(dummy_schema):
            foo: int


def test_add_const_fields_enum():
    SomeEnum = enum.Enum("SomeEnum", dict(a="A", b="B"))  # type: ignore

    class A(MetadataSchema):
        x: SomeEnum

    # works without override (restriction)
    @add_const_fields({"x": SomeEnum.a})
    class B(A):
        ...

    # invalid without override
    with pytest.raises(TypeError):
        # not enum value
        @add_const_fields({"x": "invalid"})
        class C(A):
            ...

    with pytest.raises(ValueError):
        # already defined as constant field
        @add_const_fields({"x": SomeEnum.b})
        class D(B):
            ...


def test_add_const_fields_literal():
    class A(MetadataSchema):
        x: Literal["foo", 42]

    # works without override (restriction)
    @add_const_fields({"x": "foo"})
    class B(A):
        ...

    # invalid without override
    with pytest.raises(TypeError):
        # not type value
        @add_const_fields({"x": "invalid"})
        class C(A):
            ...

    with pytest.raises(ValueError):
        # already defined as constant field
        @add_const_fields({"x": 42})
        class D(B):
            ...


def test_make_mandatory(dummy_schema):
    with pytest.raises(ValueError):
        # no such field
        @make_mandatory("bar")
        class DummyChild1(dummy_schema):
            ...

    with pytest.raises(ValueError):
        # collision: decorator + definition
        @make_mandatory("foo")
        class DummyChild2(dummy_schema):
            foo: int

    @make_mandatory("foo")
    class DummyChild(dummy_schema):
        ...

    # field is already defined by decorator
    with pytest.raises(ValueError):
        make_mandatory("foo")(DummyChild)

    # parent still ok with optional field
    assert dummy_schema().foo is None

    # child now requires it
    with pytest.raises(ValidationError):
        DummyChild()  # missing foo
    assert DummyChild(foo=5).foo == 5  # ok


def test_override(dummy_schema, plugingroups_test):
    schemas = plugingroups_test.get("schema")

    assert dummy_schema(foo=123).foo == 123

    class DummyChild(dummy_schema):
        class Plugin:
            name = "test.dummy"
            version = (0, 1, 0)

        foo: str

    assert DummyChild(foo="hello").foo == "hello"

    with pytest.raises(TypeError):
        # str not subtype of optional[int]
        register_in_group(schemas, DummyChild, violently=True)

    # with override it works
    override("foo")(DummyChild)
    register_in_group(schemas, DummyChild, violently=True)

    with pytest.raises(ValueError):
        # missing override hint for foo
        @register_in_group(schemas, violently=True)
        @override("foo")
        class DummyChild2(dummy_schema):
            class Plugin:
                name = "test.dummy2"
                version = (0, 1, 0)

    with pytest.raises(ValueError):
        # override non-existing -> error
        @register_in_group(schemas, violently=True)
        @override("bar")
        class DummyChild3(dummy_schema):
            class Plugin:
                name = "test.dummy3"
                version = (0, 1, 0)



"""This module defines a metaclass that should be used for all plugin types."""

from .types import to_semver_str


class MarkerMixin:
    """Base class for Metador-internal marker mixins.

    It can be used to hand out copies of classes with markings,
    without actually modifying the original class.
    """

    @classmethod
    def _fieldname(cls):
        return f"__{cls.__name__}_unwrapped__"

    @classmethod
    def _is_marked(cls, c) -> bool:
        """Return whether `c` is proper subclass of this marker."""
        return c is not cls and issubclass(c, cls)

    @classmethod
    def _mark_class(cls, c):
        """Mark a class with this marker mixin."""
        if cls._is_marked(c):
            raise TypeError(f"{c} already marked by {cls}!")

        ret = c.__class__(c.__name__, (cls, c), {})
        # ret.__module__ = c.__module__
        setattr(ret, cls._fieldname(), c)

        # NOTE: discouraged!
        # https://docs.python.org/3/howto/annotations.html#annotations-howto
        # ----
        # anns = getattr(ret, "__annotations__", {})
        # anns[unw_field] = ClassVar[Type]
        # ret.__annotations__ = anns

        return ret

    @classmethod
    def _unwrap(cls, c):
        """Return the original class, or None if given argument is not marked."""
        if issubclass(c, cls):
            return getattr(c, cls._fieldname())
        else:
            return None


class UndefVersion(MarkerMixin):
    """Marker for a plugin class retrieved with no specified version.

    We have to do this crazy thing, because wrapt.ObjectProxy-wrapped
    classes can be transparently derived, and what is even worse,
    the derived class is not wrapped anymore.

    The mixin subclass approach therefore makes more sense here, as
    the metaclass then can check for its presence.
    """

    @classmethod
    def _mark_class(cls, c):
        # NOTE: we also want to mark nested non-plugins to prevent subclassing
        # so we do not assume that cls.Plugin is defined
        ret = super()._mark_class(c)
        # make sure that the Plugin section *is* actually inherited,
        # (normally this is prevented by the plugin metaclass)
        # that way we can use the marked class as if it was the real one
        if not ret.__dict__.get("Plugin"):
            ret.Plugin = c.Plugin

        return ret


class PluginMetaclassMixin(type):
    """Metaclass mixin to be used with plugins of any group.

    It provides an is_plugin property to classes to quickly check if they
    seem to be valid registered plugins.

    It ensures that:
    * the `Plugin` inner class is not automatically inherited
    * registered Plugin classes cannot be subclassed if loaded without a fixed version

    For the second part, this works together with the PluginGroup implementation,
    which makes sure that schemas requested without versions are not actually handed out,
    but instead users get a subclass with the `UndefVersion` mixin we can detect here.
    """

    def __repr__(self):
        # add plugin name and version to default class repr
        if c := UndefVersion._unwrap(self):
            # indicate the attached UndefVersion
            return f"{repr(c)} (version unspecified)"
        else:
            # indicate loaded plugin name and version
            pg_str = ""
            if pgi := self.__dict__.get("Plugin"):
                pgi = self.Plugin
                pg_str = f" ({pgi.name} {to_semver_str(pgi.version)})"

            return f"{super().__repr__()}{pg_str}"

    def __new__(cls, name, bases, dct):
        # prevent inheriting from a plugin accessed without stated version
        for b in bases:
            if UndefVersion._is_marked(b):
                if pgi := b.__dict__.get("Plugin"):
                    ref = f"plugin '{pgi.name}'"
                else:
                    ref = f"{UndefVersion._unwrap(b)} originating from a plugin"
                msg = f"{name}: Cannot inherit from {ref} of unspecified version!"
                raise TypeError(msg)

        # prevent inheriting inner Plugin class by setting it to None
        if "Plugin" not in dct:
            dct["Plugin"] = None

        # hide special marker base class from parent metaclass (if present)
        # so it does not have to know about any of this happening
        # (otherwise it could interfere with other checks)
        # NOTE: needed e.g. for schemas to work properly
        filt_bases = tuple(b for b in bases if b is not UndefVersion)
        ret = super().__new__(cls, name, filt_bases, dct)

        # add marker back, as if it was present all along
        if len(filt_bases) < len(bases):
            ret.__bases__ = (UndefVersion, *ret.__bases__)

        return ret



"""IH5 skeletons and stubs (low-level structures used by IH5MFRecord).

A skeleton is documenting the tree structure of a HDF5-like container,
ignoring the actual data content (attribute values and datasets).

This can be used to implement manifest file and support "patching in thin air",
i.e. without having the actual container.
"""
from typing import Dict, Literal, Union

import h5py
from pydantic import BaseModel

from .overlay import H5Type, IH5Dataset, IH5Group
from .record import IH5Record, IH5UserBlock


class SkeletonNodeInfo(BaseModel):
    node_type: Literal[H5Type.group, H5Type.dataset]
    patch_index: int  # if its a dataset -> patch where to get data, group: creation patch
    attrs: Dict[str, int]  # names and patch indices of attributes

    def with_patch_index(self, idx: int):
        ret = self.copy()
        ret.attrs = dict(self.attrs)
        ret.patch_index = idx
        for k in ret.attrs.keys():
            ret.attrs[k] = idx
        return ret

    @classmethod
    def for_node(cls, node: Union[IH5Group, IH5Dataset]):
        if isinstance(node, IH5Dataset):
            dt = H5Type.dataset
        else:
            dt = H5Type.group

        def cidx_to_patch_idx(cidx):
            return node._record._ublock(cidx).patch_index

        pidx = cidx_to_patch_idx(node._cidx)
        ats = {
            key: cidx_to_patch_idx(node.attrs._find(key)) for key in node.attrs.keys()
        }

        return cls(node_type=dt, patch_index=pidx, attrs=ats)


class IH5Skeleton(BaseModel):
    __root__: Dict[str, SkeletonNodeInfo]

    def with_patch_index(self, idx: int):
        """Copy of skeletonwith  patch index of all nodes and attributes modified."""
        ret = self.copy()
        ret.__root__ = dict(self.__root__)
        for k, v in ret.__root__.items():
            ret.__root__[k] = v.with_patch_index(idx)
        return ret

    @classmethod
    def for_record(cls, rec: IH5Record):
        """Return mapping from all paths in a IH5 record to their type.

        The attributes are represented as special paths with the shape `a/b/.../n@attr`,
        pointing to the attribute named `attr` at the path `a/b/.../n`.

        First component is a H5Type enum value,
        Second component is more detailed type for attribute values and `IH5Dataset`s.
        """
        skel = {"/": SkeletonNodeInfo.for_node(rec["/"])}

        def add_paths(_, node):
            skel[node.name] = SkeletonNodeInfo.for_node(node)

        rec.visititems(add_paths)
        return cls(__root__=skel)


# NOTE: we pass in the empty container as first argument in the following
# in order to make this generic over subtypes (IH5Record, IH5MFRecord)!


def init_stub_skeleton(ds: IH5Record, skel: IH5Skeleton):
    """Fill a passed fresh container with stub structure based on a skeleton."""
    if len(ds) or len(ds.attrs):
        raise ValueError("Container not empty, cannot initialize stub structure here!")

    for k, v in skel.__root__.items():
        if v.node_type == H5Type.group:
            if k not in ds:
                ds.create_group(k)
        elif v.node_type == H5Type.dataset:
            ds[k] = h5py.Empty(None)

        for a in v.attrs.keys():
            ds[k].attrs[a] = h5py.Empty(None)


def init_stub_base(target: IH5Record, src_ub: IH5UserBlock, src_skel: IH5Skeleton):
    """Prepare a stub base container, given empty target, source user block and skeleton.

    Patches on top of this container will work with the original container.
    """
    init_stub_skeleton(target, src_skel)
    # mark as base container
    target._set_ublock(-1, src_ub.copy(update={"prev_patch": None}))



from __future__ import annotations

import json
from typing import List, Optional, Set, Union

import pytest
from frozendict import frozendict
from pydantic import BaseModel

from metador_core.schema.encoder import add_json_encoder
from metador_core.util.models import (
    atomic_types,
    field_atomic_types,
    field_origins,
    field_parent_type,
    new_fields,
    updated_fields,
)

add_json_encoder(frozendict, lambda x: json.dumps(dict(x)))


class MyBaseModel(BaseModel):
    foo: Optional[List[Union[int, Set[str]]]]

    class Config:
        json_encoders = {frozendict: lambda x: dict(x.items())}

    def dict(self, *args, **kwargs):
        return frozendict(super().dict(*args, **kwargs))


class ModelA(MyBaseModel):
    bar: Union[int, str]
    baz: bool


class ModelB(ModelA):
    foo: List[int]  # type: ignore


class ModelC(ModelB):
    bar: str


def test_field_origins():
    assert list(field_origins(ModelC, "foo")) == [ModelB, MyBaseModel]
    assert list(field_origins(ModelC, "bar")) == [ModelC, ModelA]
    assert list(field_origins(ModelC, "baz")) == [ModelA]
    assert list(field_origins(ModelC, "qux")) == []  # no such field -> no origins


def test_updated_fields():
    assert set(updated_fields(MyBaseModel)) == {"foo"}
    assert set(updated_fields(ModelA)) == {"bar", "baz"}
    assert set(updated_fields(ModelB)) == {"foo"}
    assert set(updated_fields(ModelC)) == {"bar"}


def test_new_fields():
    assert set(new_fields(MyBaseModel)) == {"foo"}
    assert set(new_fields(ModelA)) == {"bar", "baz"}
    assert set(new_fields(ModelB)) == set()
    assert set(new_fields(ModelC)) == set()


def test_field_atomic_types():
    # no upper bound
    assert list(field_atomic_types(MyBaseModel.__fields__["foo"])) == [int, str]
    # with upper bound
    assert list(field_atomic_types(MyBaseModel.__fields__["foo"], bound=int)) == [int]


def test_collect_model_types():
    # no upper bound
    ret = atomic_types(ModelB)
    assert "qux" not in ret
    assert ret["foo"] == {int}
    assert ret["bar"] == {int, str}
    assert ret["baz"] == {bool}

    # add a upper bound on collected types
    ret = atomic_types(ModelC, bound=int)
    assert ret["foo"] == {int}
    assert ret["bar"] == set()
    assert ret["baz"] == {bool}


def test_field_parent():
    assert field_parent_type(ModelC, "foo") == List[int]
    assert field_parent_type(ModelB, "foo") == Optional[List[Union[int, Set[str]]]]
    assert field_parent_type(ModelC, "bar") == Union[int, str]
    assert field_parent_type(ModelC, "baz") == bool

    with pytest.raises(ValueError):
        field_parent_type(ModelC, "qux")  # no hint, not defined
    with pytest.raises(ValueError):
        print(field_parent_type(ModelC, "quux"))  # no hint, but defined



from __future__ import annotations

from itertools import takewhile
from typing import Any, Dict, MutableMapping, Optional, Set, Type, Union

import h5py
import wrapt

from . import utils as M
from .drivers import MetadorDriver, to_h5filelike
from .interface import MetadorContainerTOC, MetadorMeta, NodeAcl, NodeAclFlags
from .protocols import H5DatasetLike, H5FileLike, H5GroupLike, H5NodeLike, OpenMode


class UnsupportedOperationError(AttributeError):
    """Subclass to distinguish between actually missing attribute and unsupported one."""


class WrappedAttributeManager(wrapt.ObjectProxy):
    """Wrapper for AttributeManager-like objects to prevent mutation (read-only) or inspection (skel-only)."""

    __wrapped__: MutableMapping

    _self_acl: NodeAclFlags
    _self_acl_whitelist: Dict[NodeAcl, Set[str]] = {
        NodeAcl.skel_only: {"keys"},
        NodeAcl.read_only: {"keys", "values", "items", "get"},
    }
    _self_allowed: Set[str]

    def __init__(self, obj, acl: NodeAclFlags):
        super().__init__(obj)
        # initialize whitelist based on passed ACL flags
        self._self_acl = acl
        allowed: Optional[Set[str]] = None
        for flag, value in acl.items():
            if not value or flag not in self._self_acl_whitelist:
                continue
            if allowed is None:
                allowed = self._self_acl_whitelist[flag]
            else:
                allowed = allowed.intersection(self._self_acl_whitelist[flag])
        self._self_allowed = allowed or set()

    def _raise_illegal_op(self, flag_info: str):
        raise UnsupportedOperationError(
            f"This attribute set belongs to a node marked as {flag_info}!"
        )

    def __getattr__(self, key: str):
        # NOTE: this will not restrict __contains__ because its a special method
        # (which is desired behavior).
        if (
            hasattr(self.__wrapped__, key)
            and self._self_allowed
            and key not in self._self_allowed
        ):
            self._raise_illegal_op(str(self._self_acl))
        return getattr(self.__wrapped__, key)

    # this is not intercepted by getattr
    def __getitem__(self, key: str):
        if self._self_acl[NodeAcl.skel_only]:
            self._raise_illegal_op(NodeAcl.skel_only.name)
        return self.__wrapped__.__getitem__(key)

    # this is not intercepted by getattr
    def __setitem__(self, key: str, value):
        if self._self_acl[NodeAcl.read_only]:
            self._raise_illegal_op(NodeAcl.read_only.name)
        return self.__wrapped__.__setitem__(key, value)

    # this is not intercepted by getattr
    def __delitem__(self, key: str):
        if self._self_acl[NodeAcl.read_only]:
            self._raise_illegal_op(NodeAcl.read_only.name)
        return self.__wrapped__.__delitem__(key)

    def __repr__(self) -> str:
        return repr(self.__wrapped__)


class WithDefaultQueryStartNode(wrapt.ObjectProxy):
    """Cosmetic wrapper to search metadata below a H5LikeGroup.

    Used to make sure that:
        `group.metador.query(...)`
    is equivalent to:
        `container.metador.query(..., node=group)`

    (without it, the default node will be the root "/" if not specified).
    """

    __wrapped__: MetadorContainerTOC

    def __init__(self, obj: MetadorContainerTOC, def_start_node):
        super().__init__(obj)
        self._self_query_start_node = def_start_node

    def query(self, schema, version=None, *, node=None):
        node = node or self._self_query_start_node
        return self.__wrapped__.query(schema, version, node=node)


class MetadorNode(wrapt.ObjectProxy):
    """Wrapper for h5py and IH5 Groups and Datasets providing Metador-specific features.

    In addition to the Metadata management, also provides helpers to reduce possible
    mistakes in implementing interfaces by allowing to mark nodes as

    * read_only (regardless of the writability of the underlying opened container) and
    * local_only (preventing access to (meta)data above this node)

    Note that these are "soft" restrictions to prevent errors and can be bypassed.
    """

    __wrapped__: H5NodeLike

    @staticmethod
    def _parse_access_flags(kwargs) -> NodeAclFlags:
        # NOTE: mutating kwargs, removes keys that are inspected!
        return {flag: kwargs.pop(flag.name, False) for flag in iter(NodeAcl)}

    def __init__(self, mc: MetadorContainer, node: H5NodeLike, **kwargs):
        flags = self._parse_access_flags(kwargs)
        lp = kwargs.pop("local_parent", None)
        if kwargs:
            raise ValueError(f"Unknown keyword arguments: {kwargs}")

        super().__init__(node)
        self._self_container: MetadorContainer = mc

        self._self_flags: NodeAclFlags = flags
        self._self_local_parent: Optional[MetadorGroup] = lp

    def _child_node_kwargs(self):
        """Return kwargs to be passed to a child node.

        Ensures that {read,skel,local}_only status is passed down correctly.
        """
        return {
            "local_parent": self if self.acl[NodeAcl.local_only] else None,
            **{k.name: v for k, v in self.acl.items() if v},
        }

    def restrict(self, **kwargs) -> MetadorNode:
        """Restrict this object to be local_only or read_only.

        Pass local_only=True and/or read_only=True to enable the restriction.

        local_only means that the node may not access the parent or file objects.
        read_only means that mutable actions cannot be done (even if container is mutable).
        """
        added_flags = self._parse_access_flags(kwargs)
        if added_flags[NodeAcl.local_only]:
            # was set as local explicitly for this node ->
            self._self_local_parent = None  # remove its ability to go up

        # can only set, but not unset!
        self._self_flags.update({k: True for k, v in added_flags.items() if v})
        if kwargs:
            raise ValueError(f"Unknown keyword arguments: {kwargs}")

        return self

    @property
    def acl(self) -> Dict[NodeAcl, bool]:
        """Return ACL flags of current node."""
        return dict(self._self_flags)

    def _guard_path(self, path: str):
        if M.is_internal_path(path):
            msg = f"Trying to use a Metador-internal path: '{path}'"
            raise ValueError(msg)
        if self.acl[NodeAcl.local_only] and path[0] == "/":
            msg = f"Node is marked as local_only, cannot use absolute path '{path}'!"
            raise ValueError(msg)

    def _guard_acl(self, flag: NodeAcl, method: str = "this method"):
        if self.acl[flag]:
            msg = f"Cannot use {method}, the node is marked as {flag.name}!"
            raise UnsupportedOperationError(msg)

    # helpers

    def _wrap_if_node(self, val):
        """Wrap value into a metador node wrapper, if it is a suitable group or dataset."""
        if isinstance(val, H5GroupLike):
            return MetadorGroup(self._self_container, val, **self._child_node_kwargs())
        elif isinstance(val, H5DatasetLike):
            return MetadorDataset(
                self._self_container, val, **self._child_node_kwargs()
            )
        else:
            return val

    def _destroy_meta(self, _unlink: bool = True):
        """Destroy all attached metadata at and below this node."""
        self.meta._destroy(_unlink=_unlink)

    # need that to add our new methods

    def __dir__(self):
        names = set.union(
            *map(
                lambda x: set(x.__dict__.keys()),
                takewhile(lambda x: issubclass(x, MetadorNode), type(self).mro()),
            )
        )
        return list(set(super().__dir__()).union(names))

    # make wrapper transparent

    def __repr__(self):
        return repr(self.__wrapped__)

    # added features

    @property
    def meta(self) -> MetadorMeta:
        """Access the interface to metadata attached to this node."""
        return MetadorMeta(self)

    @property
    def metador(self) -> MetadorContainerTOC:
        """Access the info about the container this node belongs to."""
        return WithDefaultQueryStartNode(self._self_container.metador, self)

    # wrap existing methods as needed

    @property
    def name(self) -> str:
        return self.__wrapped__.name  # just for type checker not to complain

    @property
    def attrs(self):
        if self.acl[NodeAcl.read_only] or self.acl[NodeAcl.skel_only]:
            return WrappedAttributeManager(self.__wrapped__.attrs, self.acl)
        return self.__wrapped__.attrs

    @property
    def parent(self) -> MetadorGroup:
        if self.acl[NodeAcl.local_only]:
            # allow child nodes of local-only nodes to go up to the marked parent
            # (or it is None, if this is the local root)
            if lp := self._self_local_parent:
                return lp
            else:
                # raise exception (illegal non-local access)
                self._guard_acl(NodeAcl.local_only, "parent")

        return MetadorGroup(
            self._self_container,
            self.__wrapped__.parent,
            **self._child_node_kwargs(),
        )

    @property
    def file(self) -> MetadorContainer:
        if self.acl[NodeAcl.local_only]:
            # raise exception (illegal non-local access)
            self._guard_acl(NodeAcl.local_only, "parent")
        return self._self_container


class MetadorDataset(MetadorNode):
    """Metador wrapper for a HDF5 Dataset."""

    __wrapped__: H5DatasetLike

    # manually assembled from public methods which h5py.Dataset provides
    _self_RO_FORBIDDEN = {"resize", "make_scale", "write_direct", "flush"}

    def __getattr__(self, key):
        if self.acl[NodeAcl.read_only] and key in self._self_RO_FORBIDDEN:
            self._guard_acl(NodeAcl.read_only, key)
        if self.acl[NodeAcl.skel_only] and key == "get":
            self._guard_acl(NodeAcl.skel_only, key)

        return getattr(self.__wrapped__, key)

    # prevent getter of node if marked as skel_only
    def __getitem__(self, *args, **kwargs):
        self._guard_acl(NodeAcl.skel_only, "__getitem__")
        return self.__wrapped__.__getitem__(*args, **kwargs)

    # prevent mutating method calls of node is marked as read_only

    def __setitem__(self, *args, **kwargs):
        self._guard_acl(NodeAcl.read_only, "__setitem__")
        return self.__wrapped__.__setitem__(*args, **kwargs)


# TODO: can this be done somehow with wrapt.decorator but still without boilerplate?
# problem is it wants a function, but we need to look it up by name first
# so we hand-roll the decorator for now.
def _wrap_method(method: str, is_read_only_method: bool = False):
    """Wrap a method called on a HDF5 entity.

    Prevents user from creating or deleting reserved entities/names by hand.
    Ensures that a wrapped Group/Dataset is returned.

    If is_read_only=False and the object is read_only, refuses to call the method.
    """

    def wrapped_method(obj, name, *args, **kwargs):
        # obj will be the self of the wrapper instance
        obj._guard_path(name)
        if not is_read_only_method:
            obj._guard_acl(NodeAcl.read_only, method)
        ret = getattr(obj.__wrapped__, method)(name, *args, **kwargs)  # RAW
        return obj._wrap_if_node(ret)

    return wrapped_method


# classes of h5py reference-like types (we don't support that)
_H5_REF_TYPES = [h5py.HardLink, h5py.SoftLink, h5py.ExternalLink, h5py.h5r.Reference]


class MetadorGroup(MetadorNode):
    """Wrapper for a HDF5 Group."""

    __wrapped__: H5GroupLike

    def _destroy_meta(self, _unlink: bool = True):
        """Destroy all attached metadata at and below this node (recursively)."""
        super()._destroy_meta(_unlink=_unlink)  # this node
        for child in self.values():  # recurse
            child._destroy_meta(_unlink=_unlink)

    # these access entities in read-only way:

    get = _wrap_method("get", is_read_only_method=True)
    __getitem__ = _wrap_method("__getitem__", is_read_only_method=True)

    # these just create new entities with no metadata attached:

    create_group = _wrap_method("create_group")
    require_group = _wrap_method("require_group")
    create_dataset = _wrap_method("create_dataset")
    require_dataset = _wrap_method("require_dataset")

    def __setitem__(self, name, value):
        if any(map(lambda x: isinstance(value, x), _H5_REF_TYPES)):
            raise ValueError(f"Unsupported reference type: {type(value).__name__}")

        return _wrap_method("__setitem__")(self, name, value)

    # following all must be filtered to hide metador-specific structures:

    # must wrap nodes passed into the callback function and filter visited names
    def visititems(self, func):
        def wrapped_func(name, node):
            if M.is_internal_path(node.name):
                return  # skip path/node
            return func(name, self._wrap_if_node(node))

        return self.__wrapped__.visititems(wrapped_func)  # RAW

    # paths passed to visit also must be filtered, so must override this one too
    def visit(self, func):
        def wrapped_func(name, _):
            return func(name)

        return self.visititems(wrapped_func)

    # following also depend on the filtered sequence, directly
    # filter the items, derive other related functions based on that

    def items(self):
        for k, v in self.__wrapped__.items():
            if v is None:
                # NOTE: e.g. when nodes are deleted/moved during iteration,
                # v can suddenly be None -> we need to catch this case!
                continue
            if not M.is_internal_path(v.name):
                yield (k, self._wrap_if_node(v))

    def values(self):
        return map(lambda x: x[1], self.items())

    def keys(self):
        return map(lambda x: x[0], self.items())

    def __iter__(self):
        return iter(self.keys())

    def __len__(self):
        return len(list(self.keys()))

    def __contains__(self, name: str):
        self._guard_path(name)
        if name[0] == "/" and self.name != "/":
            return name in self["/"]
        segs = name.lstrip("/").split("/")
        has_first_seg = segs[0] in self.keys()
        if len(segs) == 1:
            return has_first_seg
        else:
            if nxt := self.get(segs[0]):
                return "/".join(segs[1:]) in nxt
            return False

    # these we can take care of but are a bit more tricky to think through

    def __delitem__(self, name: str):
        self._guard_acl(NodeAcl.read_only, "__delitem__")
        self._guard_path(name)

        node = self[name]
        # clean up metadata (recursively, if a group)
        node._destroy_meta()
        # kill the actual data
        return _wrap_method("__delitem__")(self, name)

    def move(self, source: str, dest: str):
        self._guard_acl(NodeAcl.read_only, "move")
        self._guard_path(source)
        self._guard_path(dest)

        src_metadir = self[source].meta._base_dir
        # if actual data move fails, an exception will prevent the rest
        self.__wrapped__.move(source, dest)  # RAW

        # if we're here, no problems -> proceed with moving metadata
        dst_node = self[dest]
        if isinstance(dst_node, MetadorDataset):
            dst_metadir = dst_node.meta._base_dir
            # dataset has its metadata stored in parallel -> need to take care of it
            meta_base = dst_metadir
            if src_metadir in self.__wrapped__:  # RAW
                self.__wrapped__.move(src_metadir, dst_metadir)  # RAW
        else:
            # directory where to fix up metadata object TOC links
            # when a group was moved, all metadata is contained in dest -> search it
            meta_base = dst_node.name

        # re-link metadata object TOC links
        if meta_base_node := self.__wrapped__.get(meta_base):
            assert isinstance(meta_base_node, H5GroupLike)
            missing = self._self_container.metador._links.find_missing(meta_base_node)
            self._self_container.metador._links.repair_missing(missing, update=True)

    def copy(
        self,
        source: Union[str, MetadorGroup, MetadorDataset],
        dest: Union[str, MetadorGroup],
        **kwargs,
    ):
        self._guard_acl(NodeAcl.read_only, "copy")

        # get source node and its name without the path and its type
        src_node: MetadorNode
        if isinstance(source, str):
            self._guard_path(source)
            src_node = self[source]
        elif isinstance(source, MetadorNode):
            src_node = source
        else:
            raise ValueError("Copy source must be path, Group or Dataset!")
        src_is_dataset: bool = isinstance(src_node, MetadorDataset)
        src_name: str = src_node.name.split("/")[-1]
        # user can override name at target
        dst_name: str = kwargs.pop("name", src_name)

        # fix up target path
        dst_path: str
        if isinstance(dest, str):
            self._guard_path(dest)
            dst_path = dest
        elif isinstance(dest, MetadorGroup):
            dst_path = dest.name + f"/{dst_name}"
        else:
            raise ValueError("Copy dest must be path or Group!")

        # get other allowed options
        without_attrs: bool = kwargs.pop("without_attrs", False)
        without_meta: bool = kwargs.pop("without_meta", False)
        if kwargs:
            raise ValueError(f"Unknown keyword arguments: {kwargs}")

        # perform copy
        copy_kwargs = {
            "name": None,
            "shallow": False,
            "expand_soft": True,
            "expand_external": True,
            "expand_refs": True,
            "without_attrs": without_attrs,
        }
        self.__wrapped__.copy(source, dst_path, **copy_kwargs)  # RAW
        dst_node = self[dst_path]  # exists now

        if src_is_dataset and not without_meta:
            # because metadata lives in parallel group, need to copy separately:
            src_meta: str = src_node.meta._base_dir
            dst_meta: str = dst_node.meta._base_dir  # node will not exist yet
            self.__wrapped__.copy(src_meta, dst_meta, **copy_kwargs)  # RAW

            # register in TOC:
            dst_meta_node = self.__wrapped__[dst_meta]
            assert isinstance(dst_meta_node, H5GroupLike)
            missing = self._self_container.metador._links.find_missing(dst_meta_node)
            self._self_container.metador._links.repair_missing(missing)

        if not src_is_dataset:
            if without_meta:
                # need to destroy copied metadata copied with the source group
                # but keep TOC links (they point to original copy!)
                dst_node._destroy_meta(_unlink=False)
            else:
                # register copied metadata objects under new uuids
                missing = self._self_container.metador._links.find_missing(dst_node)
                self._self_container.metador._links.repair_missing(missing)

    def __getattr__(self, key):
        if hasattr(self.__wrapped__, key):
            raise UnsupportedOperationError(key)  # deliberately unsupported
        else:
            msg = f"'{type(self).__name__}' object has no attribute '{key}'"
            raise AttributeError(msg)


# ----


class MetadorContainer(MetadorGroup):
    """Wrapper class adding Metador container interface to h5py.File-like objects.

    The wrapper ensures that any actions done to IH5Records through this interface
    also work with plain h5py.Files.

    There are no guarantees about behaviour with h5py methods not supported by IH5Records.

    Given `old: MetadorContainer`, `MetadorContainer(old.data_source, driver=old.data_driver)`
    should be able to construct another object to access the same data (assuming it is not locked).
    """

    __wrapped__: H5FileLike

    _self_SUPPORTED: Set[str] = {"mode", "flush", "close"}

    # ---- new container-level interface ----

    _self_toc: MetadorContainerTOC

    @property
    def metador(self) -> MetadorContainerTOC:
        """Access interface to Metador metadata object index."""
        return self._self_toc

    def __init__(
        self,
        name_or_obj: Union[MetadorDriver, Any],
        mode: Optional[OpenMode] = "r",
        *,
        # NOTE: driver takes class instead of enum to also allow subclasses
        driver: Optional[Type[MetadorDriver]] = None,
    ):
        """Initialize a MetadorContainer instance from file(s) or a supported object.

        The `mode` argument is ignored when simply wrapping an object.

        If a data source such as a path is passed, will instantiate the object first,
        using the default H5File driver or the passed `driver` keyword argument.
        """
        # wrap the h5file-like object (will set self.__wrapped__)
        super().__init__(self, to_h5filelike(name_or_obj, mode, driver=driver))
        # initialize metador-specific stuff
        self._self_toc = MetadorContainerTOC(self)

    # not clear if we want these in the public interface. keep this private for now:

    # def _find_orphan_meta(self) -> List[str]:
    #     """Return list of paths to metadata that has no corresponding user node anymore."""
    #     ret: List[str] = []

    #     def collect_orphans(name: str):
    #         if M.is_meta_base_path(name):
    #             if M.to_data_node_path(name) not in self:
    #                 ret.append(name)

    #     self.__wrapped__.visit(collect_orphans)
    #     return ret

    # def _repair(self, remove_orphans: bool = False):
    #     """Repair container structure on best-effort basis.

    #     This will ensure that the TOC points to existing metadata objects
    #     and that all metadata objects are listed in the TOC.

    #     If remove_orphans is set, will erase metadata not belonging to an existing node.

    #     Notice that missing schema plugin dependency metadata cannot be restored.
    #     """
    #     if remove_orphans:
    #         for path in self._find_orphan_meta():
    #             del self.__wrapped__[path]
    #     self.toc._links.find_broken(repair=True)
    #     missing = self.toc._links._find_missing("/")
    #     self.toc._links.repair_missing(missing)

    # ---- pass through HDF5 group methods to a wrapped root group instance ----

    def __getattr__(self, key: str):
        if key in self._self_SUPPORTED:
            return getattr(self.__wrapped__, key)
        return super().__getattr__(key)  # ask group for method

    # context manager: return the wrapper back, not the raw thing:

    def __enter__(self):
        self.__wrapped__.__enter__()
        return self

    def __exit__(self, *args):
        return self.__wrapped__.__exit__(*args)

    # we want these also to be forwarded to the wrapped group, not the raw object:

    def __dir__(self):
        return list(set(super().__dir__()).union(type(self).__dict__.keys()))

    # make wrapper transparent:

    def __repr__(self) -> str:
        return repr(self.__wrapped__)  # shows that its a File, not just a Group



"""Hacks to improve pydantic JSON Schema generation."""

import json
from functools import partial
from typing import Any, Dict, Iterable, List, Type, Union

from pydantic import BaseModel
from pydantic import schema_of as pyd_schema_of
from pydantic.schema import schema as pyd_schemas
from typing_extensions import TypeAlias

from ..util.hashsums import hashsum
from ..util.models import updated_fields

KEY_SCHEMA_DEFS = "$defs"
"""JSON schema key to store subschema definitions."""

KEY_SCHEMA_HASH = "$jsonschema_hash"
"""Custom key to store schema hashsum."""

# ----

JSON_PRIMITIVE_TYPES = (type(None), bool, int, float, str)

# shallow type definitions
JSONPrimitive: TypeAlias = Union[None, bool, int, float, str]
JSONObject: TypeAlias = Dict[str, Any]
JSONArray: TypeAlias = List[Any]
JSONType: TypeAlias = Union[JSONPrimitive, JSONArray, JSONObject]

JSONSCHEMA_STRIP = {
    # these are meta-level keys that should not affect the hash:
    "title",
    "description",
    "examples",
    "$comment",
    "readOnly",
    "writeOnly",
    "deprecated",
    "$id",
    # subschemas have their own hashes, so if referenced schemas
    # change, then the $refs change automatically.
    "definitions",
    KEY_SCHEMA_DEFS,
    # we can't hash the hash, it cannot be part of the clean schema.
    KEY_SCHEMA_HASH,
}
"""Fields to be removed for JSON Schema hashsum computation."""


def clean_jsonschema(obj: JSONType, *, _is_properties: bool = False):
    if isinstance(obj, JSON_PRIMITIVE_TYPES):
        return obj
    if isinstance(obj, list):
        return list(map(clean_jsonschema, obj))
    if isinstance(obj, dict):
        return {
            k: clean_jsonschema(v, _is_properties=k == "properties")
            for k, v in obj.items()
            # must ensure not to touch keys in a properties sub-object!
            if _is_properties or k not in JSONSCHEMA_STRIP
        }

    raise ValueError(f"Object {obj} not of a JSON type: {type(obj)}")


def normalized_json(obj: JSONType) -> bytes:
    return json.dumps(
        obj,
        ensure_ascii=True,
        allow_nan=False,
        indent=None,
        sort_keys=True,
        separators=(",", ":"),
    ).encode("utf-8")


def jsonschema_id(schema: JSONType):
    """Compute robust semantic schema identifier.

    A schema identifier is based on the schema plugin name + version
    and its JSON Schema representation, which includes all parent and nested schemas.
    """
    return hashsum(normalized_json(clean_jsonschema(schema)), "sha256")


# ----


def lift_nested_defs(schema: JSONObject):
    """Flatten nested $defs ($defs -> key -> $defs) in-place."""
    if mydefs := schema.get(KEY_SCHEMA_DEFS):
        inner = []
        for schema in mydefs.values():
            lift_nested_defs(schema)
            if nested := schema.pop(KEY_SCHEMA_DEFS, None):
                inner.append(nested)
        for nested in inner:
            mydefs.update(nested)


KEY_PYD_DEFS = "definitions"
"""key name where pydantic stores subschema definitions."""

REF_PREFIX = f"#/{KEY_PYD_DEFS}/"
"""default $refs prefix of pydantic."""


def merge_nested_defs(schema: JSONObject):
    """Merge definitions in-place."""
    if defs := schema.pop(KEY_PYD_DEFS, None):
        my_defs = schema.get(KEY_SCHEMA_DEFS)
        if not my_defs:
            schema[KEY_SCHEMA_DEFS] = {}
            my_defs = schema[KEY_SCHEMA_DEFS]
        # update, by preserve existing
        defs.update(my_defs)
        my_defs.update(defs)


# ----


def collect_defmap(defs: JSONObject):
    """Compute dict mapping current name in $defs to new name based on metador_hash."""
    defmap = {}
    for name, subschema in defs.items():
        if KEY_SCHEMA_HASH in subschema:
            defmap[name] = subschema[KEY_SCHEMA_HASH].strip("/")
        else:
            # print("no hashsum: ", name)
            defmap[name] = name

    return defmap


def map_ref(defmap, refstr: str):
    """Update the `$ref` string based on defmap.

    Will replace `#/definitions/orig`
    with `#/$defs/mapped`.
    """
    if refstr.startswith(REF_PREFIX):
        # print("remap", refstr)
        plen = len(REF_PREFIX)
        if new_name := defmap.get(refstr[plen:]):
            return f"#/{KEY_SCHEMA_DEFS}/{new_name}"
    return refstr


def update_refs(defmap, obj):
    """Recursively update `$ref` in `obj` based on defmap."""
    print("update", obj)
    if isinstance(obj, (type(None), bool, int, float, str)):
        return obj
    elif isinstance(obj, list):
        return list(map(partial(update_refs, defmap), obj))
    elif isinstance(obj, dict):
        return {
            k: (update_refs(defmap, v) if k != "$ref" else map_ref(defmap, v))
            for k, v in obj.items()
        }
    raise ValueError(f"Object {obj} not of a JSON type: {type(obj)}")


def remap_refs(schema):
    """Remap the $refs to use metador_hash-based keys.

    Input must be a completed schema with a global `$defs` section
    that all nested entities use for local references.
    """
    defs = schema.pop(KEY_SCHEMA_DEFS, None)
    if not defs:  # nothing to do
        return schema

    # get name map, old -> new
    defmap = collect_defmap(defs)
    # update refs
    defs.update(update_refs(defmap, defs))
    schema.update(update_refs(defmap, schema))
    # rename defs
    schema[KEY_SCHEMA_DEFS] = {defmap[k]: v for k, v in defs.items()}


# ----


def fixup_jsonschema(schema):
    merge_nested_defs(schema)  # move `definitions` into `$defs`
    lift_nested_defs(schema)  # move nested `$defs` to top level `$defs`
    remap_refs(schema)  # "rename" defs from model name to metador hashsum


def schema_of(model: Type[BaseModel], *args, **kwargs):
    """Return JSON Schema for a model.

    Improved version of `pydantic.schema_of`, returns result
    in $defs normal form, with $ref pointing to the model.
    """
    schema = pyd_schema_of(model, *args, **kwargs)
    print(type(schema), schema)
    schema.pop("title", None)
    fixup_jsonschema(schema)
    return schema


def schemas(models: Iterable[Type[BaseModel]], *args, **kwargs):
    """Return JSON Schema for multiple models.

    Improved version of `pydantic.schema.schema`,
    returns result in $defs normal form.
    """
    schema = pyd_schemas(tuple(models), *args, **kwargs)
    fixup_jsonschema(schema)
    return schema


# ----


def split_model_inheritance(schema: JSONObject, model: Type[BaseModel]):
    """Decompose a model into an allOf combination with a parent model.

    This is ugly because pydantic does in-place wrangling and caching,
    and we need to hack around it.
    """
    # NOTE: important - we assume to get the $defs standard form
    # print("want schema of", model.__base__.__name__)
    base_schema = model.__base__.schema()  # type: ignore

    # compute filtered properties / required section
    schema_new = dict(schema)
    ps = schema_new.pop("properties", None)
    rq = schema_new.pop("required", None)

    lst_fields = updated_fields(model)
    ps_new = {k: v for k, v in ps.items() if k in lst_fields}
    rq_new = None if not rq else [k for k in rq if k in ps_new]
    schema_this = {k: v for k, v in [("properties", ps_new), ("required", rq_new)] if v}

    # construct new schema as combination of base schema and remainder schema
    schema_new.update(
        {
            # "rdfs:subClassOf": f"/{base_id}",
            "allOf": [{"$ref": base_schema["$ref"]}, schema_this],
        }
    )

    # we need to add the definitions to/from the base schema as well
    if KEY_SCHEMA_DEFS not in schema_new:
        schema_new[KEY_SCHEMA_DEFS] = {}
    schema_new[KEY_SCHEMA_DEFS].update(base_schema.get(KEY_SCHEMA_DEFS, {}))

    schema.clear()
    schema.update(schema_new)


def finalize_schema_extra(
    schema: JSONObject,
    model: Type[BaseModel],
    *,
    base_model: Type[BaseModel] = None,
) -> None:
    """Perform custom JSON Schema postprocessing.

    To be called as last action in custom schema_extra method in the used base model.

    Arguments:
        schema: The JSON object containing the schema
        model: The underlying pydantic model
        base_model: The custom base model that this function is called for.
    """
    base_model = base_model or BaseModel
    assert issubclass(model, base_model)

    # a schema should have a specified standard
    schema["$schema"] = "https://json-schema.org/draft/2020-12/schema"

    if model.__base__ is not base_model:
        # tricky part: de-duplicate fields from parent class
        split_model_inheritance(schema, model)

    # do this last, because it needs everything else to compute the correct hashsum:
    schema[KEY_SCHEMA_HASH] = f"{jsonschema_id(schema)}"
    fixup_jsonschema(schema)



"""This is an example packer plugin for simple general data types.

A packer plugin implements use-case specific container-related
functionality for Metador containers.

To develop your own packer plugin, implement a class deriving from
`Packer` and register the class as an entrypoint of your package
(see the `pyproject.toml` of this package, where `GenericPacker`
is registered as a packer plugin called `example`.)
"""

from pathlib import Path
from typing import Any, Union

import pandas
from overrides import overrides

from ..util.diff import DiffNode, DirDiff
from . import MetadorContainer, Packer
from .utils import DirValidationErrors, check_metadata_file, embed_file

BibMeta = Any
TableMeta = Any


class GenericPacker(Packer):
    """The generic packer is demonstrating how a packer can be implemented.

    It will pack CSV tables with metadata into corresponding HDF5 containers,
    and it will pack all other kinds of files as embedded opaque blobs.

    Both kinds of nodes will have corresponding metadata attributes attached.

    The directory is expected to have a _meta.yaml file in the container root
    and each CSV file file.csv needs a companion metadata file file.csv_meta.yaml.

    All symlinks inside the directory are completely ignored.

    This packer does very verbose logging for didactic purposes.
    Other packers may log their actions as they deem appropriate.
    """

    class Plugin:
        name = "core.generic"
        version = (0, 1, 0)

    META_SUFFIX: str = "_meta.yaml"

    @classmethod
    def sidecar_for(cls, path: Union[Path, str]) -> str:
        """Sidecar file name for given path."""
        return f"{path}{cls.META_SUFFIX}"

    @classmethod
    @overrides
    def check_dir(cls, data_dir: Path) -> DirValidationErrors:
        print("--------")
        print("called check_dir")
        errs = DirValidationErrors()
        errs.update(
            check_metadata_file(
                data_dir / cls.META_SUFFIX, required=True, schema=BibMeta
            )
        )
        return errs

    @classmethod
    @overrides
    def update(cls, mc: MetadorContainer, data_dir: Path, diff: DirDiff):
        print("--------")
        print("called update")

        for path, dnode in diff.annotate(data_dir).items():
            # the status indicates whether the file was added, removed or modified
            status = diff.status(dnode)
            print(status.value, path)

            if dnode is None:  # unchanged paths in the directory have no diff node
                print("IGNORE:", path, "(unchanged)")
                continue  # nothing to do

            if path.is_symlink():  # we ignore symlinks in the data directory
                print("IGNORE:", path, "(symlink)")
                continue

            if path.name.lower().endswith(".csv_meta.yaml"):
                # will be taken care of when the CSV file is processed
                print("IGNORE:", path, "(sidecar file)")
                continue

            # for this simple packer, each file maps 1-to-1 to a container path
            key = f"{dnode.path}"  # path inside the container

            if status == DiffNode.Status.removed:  # entity was removed ->
                # also remove in container, if it was not a symlink (which we ignored)
                if dnode.prev_type != DiffNode.ObjType.symlink:
                    print("DELETE:", key)
                    del mc[key]
                continue

            if status == DiffNode.Status.modified:  # changed
                if dnode.prev_type == dnode.curr_type and path.is_dir():
                    continue  # a changed dir should already exist + remain in container

                # otherwise it was replaced either file -> dir or dir -> file, so
                # remove entity, proceeding with loop body to add new entity version
                print("DELETE:", key)
                del mc[key]

            # now we (re-)add new or modified entities:
            if path.is_dir():
                print("CREATE:", path, "->", key, "(dir)")

                mc.create_group(key)

            elif path.is_file():
                if path.name.endswith(cls.META_SUFFIX):
                    if key == cls.META_SUFFIX:
                        # update root meta
                        print("CREATE:", path, "->", key, "(biblio metadata)")
                        mc.meta["common_biblio"] = BibMeta.parse_file(path)
                else:
                    if path.name.lower().endswith(".csv"):
                        # embed CSV as numpy array with table metadata
                        print("CREATE:", path, "->", key, "(table)")

                        mc[key] = pandas.read_csv(path).to_numpy()  # type: ignore
                        mc[key].meta["common_table"] = TableMeta.for_file(
                            cls.sidecar_for(path)
                        )

                    elif path.name.lower().endswith((".jpg", ".jpeg", ".png")):
                        # embed image file with image-specific metadata
                        print("CREATE:", path, "->", key, "(image)")
                        embed_file(mc, key, path)
                        # mc[key].meta["common_image"] = image_meta_for(path)

                    else:
                        # treat as opaque blob and add file metadata
                        print("CREATE:", path, "->", key, "(file)")
                        embed_file(mc, key, path)

                    # mc[key].meta["common_file"] = file_meta_for(path)



import isodate
import pytest
from hypothesis import assume, given
from hypothesis import strategies as st
from pydantic import BaseModel, ValidationError, parse_obj_as

import metador_core.schema.types as t
from metador_core.util.pytest import parameters

# ----
# Test our provided hints for default primitive types behave as they should


@pytest.mark.parametrize(
    "typ,val",
    parameters(
        {
            t.Bool: [0, 1, -1, 2, 0.0, 1.0, 2.1],
            t.Int: [True, False, 0.0, 1.0, 2.1],
            t.Float: [True, False, -1, 0, 1, 2],
        }
    ),
)
def test_strict_numeric_reject(typ, val):
    # make sure there is no type coercion between numeric(-like) types
    with pytest.raises(ValidationError):
        parse_obj_as(typ, val)


P_INF = float("inf")
N_INF = float("-inf")


@given(data=st.data())
@pytest.mark.parametrize(
    "t, h",
    [(bool, t.Bool), (int, t.Int), (float, t.Float), (bytes, t.Bytes), (str, t.Str)],
)
def test_def_types_accept(t, h, data):
    strategy = st.from_type(h)  # hint must be hypothesis compatible

    x = data.draw(strategy)
    assume(x == x and x != P_INF and x != N_INF)

    assert isinstance(x, t)  # result is plain type
    assert h(x) == x  # type hint is pass through
    assert parse_obj_as(h, x) == x  # works for parsing


# ----


@pytest.mark.parametrize("x", ["foo", " \tbar "])
def test_nonemptystr_accept(x):
    parse_obj_as(t.NonEmptyStr, x)


@pytest.mark.parametrize("x", ["", "\t", "  "])
def test_nonemptystr_reject(x):
    with pytest.raises(ValidationError):
        parse_obj_as(t.NonEmptyStr, x)


@pytest.mark.parametrize("x", ["application/json", "application/JSON;q=0.9;v=abc"])
def test_mimetypestr_accept(x):
    parse_obj_as(t.MimeTypeStr, x)


@pytest.mark.parametrize("x", ["invalid/mime/type", "invalid mime", "invalidMime"])
def test_mimetypestr_reject(x):
    with pytest.raises(ValidationError):
        parse_obj_as(t.MimeTypeStr, x)


@pytest.mark.parametrize("x", ["sha256:aebf", "sha512:aebf"])
def test_qualhashsumstr_accept(x):
    parse_obj_as(t.QualHashsumStr, x)


@pytest.mark.parametrize(
    "x", ["wrong:aebf", "sha512:invalid", "sha256", "sha256:", "aebf", ":aebf"]
)
def test_qualhashsumstr_reject(x):
    with pytest.raises(ValidationError):
        parse_obj_as(t.QualHashsumStr, x)


# ----


@pytest.mark.parametrize("x", [123, True])
def test_duration_reject(x):
    with pytest.raises(ValidationError):
        parse_obj_as(t.Duration, x)


@pytest.mark.parametrize("x", ["PT3H4M1S", "P3Y6M4DT12H30M5S"])
def test_duration_accept(x):
    assert isinstance(parse_obj_as(t.Duration, x), isodate.Duration)


# ----


@pytest.mark.parametrize(
    "t_cls,x",
    parameters(
        {
            t.PintUnit: [
                "meter / (second * kg) ** 2",
                "dimensionless",
                t.PintUnit("second"),
                "1",
            ],
            t.PintQuantity: [
                "123 meter / (second * kg) ** 2",
                "5",
                t.PintQuantity("7 seconds"),
            ],
        }
    ),
)
def test_pint_accept(t_cls, x):
    parse_obj_as(t_cls, x)


@pytest.mark.parametrize(
    "t_cls,x",
    parameters(
        {
            t.PintUnit: ["invalid", "2", "", 123],
            t.PintQuantity: ["123 bla", 23.12, False],
        }
    ),
)
def test_pint_reject(t_cls, x):
    with pytest.raises(ValidationError):
        parse_obj_as(t_cls, x)


@pytest.mark.parametrize(
    "cls,val", [(t.PintUnit, "meters * second"), (t.PintQuantity, "5 meters * second")]
)
def test_pint_schema_description(cls, val):
    # make sure the description is attached to jsonschema

    class SomeModel(BaseModel):
        u: cls

    assert (
        SomeModel(u=val).schema()["properties"]["u"]["title"].lower().find("pint") >= 0
    )


# from metador_core.ih5.record import IH5Record
# from metador_core.packer.util import MetadorValidationErrors
# from metador_core.schema.common import FileMeta
#
# @pytest.mark.skip(reason="FIXME port to equivalent tests in new (packer) API")
# def test_filemeta(tmp_path, tmp_ds_path):
#     # get metadata for a file, save to yaml, load from yaml, check it
#     file = tmp_path / "myfile.txt"
#     with open(file, "w") as f:
#         f.write("hello world!")

#     with open(tmp_path / "myfile_meta.yaml", "w") as f:
#         f.write(FileMeta.for_file(file).yaml())

#     # failure gives MetadorValidationErrors
#     with pytest.raises(MetadorValidationErrors):
#         FileMeta.from_file(tmp_path / "non-existing.yaml")

#     # check success
#     fmeta = FileMeta.from_file(tmp_path / "myfile_meta.yaml")
#     assert isinstance(fmeta, FileMeta)
#     assert fmeta.filename == "myfile.txt"
#     assert (
#         fmeta.hashsum
#         == "sha256:7509e5bda0c762d2bac7f90d758b5b2263fa01ccbc542ab5e3df163be08e6ca9"
#     )
#     assert fmeta.mimetype == "text/plain"

#     # try with reading from a record
#     with IH5Record(tmp_ds_path, "w") as ds:
#         ds.create_dataset("group/dataset", data=fmeta.json())
#         ds["group/dataset"].attrs["attr"] = fmeta.json()
#         ds["invalid"] = "--"
#         ds["incomplete"] = "{}"

#         with pytest.raises(MetadorValidationErrors) as e:
#             FileMeta.from_record(ds, "group")
#         assert e.value.errors["group"][0].lower().find("group") >= 0

#         with pytest.raises(MetadorValidationErrors) as e:
#             FileMeta.from_record(ds, "not-existing")
#         assert e.value.errors["not-existing"][0].lower().find("not found") >= 0

#         with pytest.raises(MetadorValidationErrors) as e:
#             FileMeta.from_record(ds, "invalid")
#         assert e.value.errors["invalid"][0].lower().find("cannot parse") >= 0

#         with pytest.raises(MetadorValidationErrors) as e:
#             FileMeta.from_record(ds, "incomplete")
#         assert e.value.errors["incomplete"][0].lower().find("missing") >= 0



"""Protocol roughly formalizing the overlap of h5py.File and IH5Record API.

We build the MetadorContainer interface assuming only these methods.
"""
from __future__ import annotations

from typing import (
    Any,
    Callable,
    ItemsView,
    Iterable,
    KeysView,
    Literal,
    MutableMapping,
    Optional,
    Protocol,
    TypeVar,
    Union,
    ValuesView,
    runtime_checkable,
)

from typing_extensions import get_args


@runtime_checkable
class H5NodeLike(Protocol):  # pragma: no cover
    """HDF5 Files, Groups and Datasets are all Nodes."""

    @property
    def name(self) -> str:
        """Absolute path of the node."""

    @property
    def attrs(self) -> MutableMapping:
        """Attached HDF5 attributes."""

    @property
    def parent(self) -> H5GroupLike:
        """Parent group."""

    @property
    def file(self) -> H5FileLike:
        """Original file-like object this node belongs to."""


@runtime_checkable
class H5DatasetLike(H5NodeLike, Protocol):  # pragma: no cover
    """Datasets provide numpy-style indexing into data.

    Metador containers use it for storing bytes,
    and for getting bytes out again using [()].
    """

    def __getitem__(self, key: Any) -> Any:
        ...

    def __setitem__(self, key: Any, value) -> None:
        ...

    # needed to distinguish from other types:
    @property
    def ndim(self) -> int:
        """Numpy-style dimensionality."""


CallbackResult = TypeVar("CallbackResult")
VisitCallback = Callable[[str], Optional[CallbackResult]]
VisititemsCallback = Callable[[str, H5NodeLike], Optional[CallbackResult]]


@runtime_checkable
class H5GroupLike(H5NodeLike, Protocol):  # pragma: no cover
    # MutableMapping-like

    def __getitem__(self, name: str) -> Union[H5GroupLike, H5DatasetLike]:
        ...

    def __setitem__(self, name: str, obj) -> None:
        ...

    def __delitem__(self, name: str) -> None:
        ...

    def __iter__(self) -> Iterable[str]:
        ...

    def __len__(self) -> int:
        ...

    # Container-like

    def __contains__(self, name: str) -> bool:
        ...

    # dict-like extras

    def keys(self) -> KeysView[str]:
        ...

    def values(self) -> ValuesView[Union[H5GroupLike, H5DatasetLike]]:
        ...

    def items(self) -> ItemsView[str, Union[H5GroupLike, H5DatasetLike]]:
        ...

    def get(self, name: str, *args, **kwargs) -> Union[H5GroupLike, H5DatasetLike]:
        ...

    # h5py specific

    def visititems(self, func: VisititemsCallback) -> Optional[CallbackResult]:
        # returns value returned from callback
        ...

    def visit(self, func: VisitCallback) -> Optional[CallbackResult]:
        # returns value returned from callback
        ...

    def create_dataset(self, path, *args, **kwargs) -> H5DatasetLike:
        # returns original passed-in data
        ...

    def require_dataset(self, path, *args, **kwargs) -> H5DatasetLike:
        ...

    def create_group(self, path: str) -> H5GroupLike:
        # returns new group
        ...

    def require_group(self, path: str) -> H5GroupLike:
        ...

    def move(self, source: str, dest: str) -> None:
        ...

    def copy(self, source, dest, **kwargs) -> None:
        ...


OpenMode = Literal["r", "r+", "a", "w", "w-", "x"]
"""User open modes that can be passed during initialization."""
_OPEN_MODES = list(get_args(OpenMode))


@runtime_checkable
class H5FileLike(H5GroupLike, Protocol):  # pragma: no cover
    """A HDF5 File acts like the root group and has some extra features."""

    @property
    def mode(self) -> Literal["r", "r+"]:
        """Return 'r' if container is immutable, otherwise 'r+'."""

    def close(self) -> None:
        ...

    # context manager (`with` notation)

    def __enter__(self) -> H5FileLike:
        ...

    def __exit__(self, ex_type, ex_value, ex_traceback) -> None:
        ...


__all__ = ["H5FileLike", "H5GroupLike", "H5DatasetLike", "H5NodeLike", "OpenMode"]



from enum import Enum
from typing import Optional

import pytest
from pydantic import Field, ValidationError
from typing_extensions import Annotated

from metador_core.schema.base import BaseModelPlus


class DummyEnum(str, Enum):
    a = "a"
    b = "b"


def test_base_model():
    with pytest.raises(ValidationError):
        # defaults are validated
        class WrongDummy(BaseModelPlus):
            foo: int = "invalid"

        WrongDummy()  # default is used -> invalid

    # now a valid model:
    class DummyModel(BaseModelPlus):
        _priv: str

        foo: Annotated[Optional[int], Field(alias="@foo")]
        bar: Optional[float]
        val: DummyEnum

    # inf/nan are forbidden
    with pytest.raises(ValidationError):
        DummyModel(bar=float("nan"))
    with pytest.raises(ValidationError):
        DummyModel(bar=float("inf"))

    # allows using non-alias name (foo) for assignment
    obj = DummyModel(foo=5, val="a", xtra="hi")

    # setters are validated
    with pytest.raises(ValidationError):
        obj.bar = "wrong"
    obj.bar = 2.41
    obj.bar = None

    # private can be set (but is ignored)
    obj._priv = "something"

    # extra is preserved on load
    assert obj.xtra == "hi"

    # yaml dump/load works
    assert DummyModel.parse_raw(obj.yaml()) == obj
    # bytes serialization (utf-8 json with newline)
    assert DummyModel.parse_raw(bytes(obj)) == obj
    # string is json with newlines
    assert DummyModel.parse_raw(str(obj)) == obj
    assert len(str(obj).split()) > 1

    # normal serializations
    assert DummyModel.parse_raw(obj.json()) == obj
    assert DummyModel.parse_obj(obj.dict()) == obj
    assert DummyModel.parse_obj(obj.json_dict()) == obj

    # test the default behavior
    js = obj.dict()

    # private fields are removed on dump
    assert "_priv" not in js

    # extra fields are preserved on dump
    assert "xtra" in js

    # enums become plain types
    assert isinstance(js["val"], str)

    # serialization to alias name
    assert "@foo" in js
    assert "foo" not in js

    # no None values
    assert "bar" not in js



"""Simplify creation of custom parsers for pydantic models."""
from typing import Any, ClassVar, Dict, Type, TypeVar

from pydantic import BaseModel

T = TypeVar("T")


class BaseParser:
    """Parsers that work with the ParserMixin must inherit from this class."""

    schema_info: Dict[str, Any] = {}
    strict: bool = True

    @classmethod
    def parse(cls, target: Type[T], v: Any) -> T:
        """Override and implement this method for custom parsing.

        The default implementation will simply pass through
        any instances of `tcls` unchanged and fail on anything else.

        Make sure that the parser can also handle any object that itself
        produces as an input.

        By default, parsers are expected to normalize the input,
        i.e. produce an instance of `tcls`, any other returned type
        will lead to an exception.

        If you know what you are doing, set `strict=False` to
        disable this behavior.

        Args:
            target: class the value should be parsed into
            v: value to be parsed
        """
        if target is not None and not isinstance(v, target):
            raise TypeError(f"Expected {target.__name__}, but got {type(v).__name__}!")
        return v


def run_parser(cls: Type[BaseParser], target: Type[T], value: Any):
    """Parse and validate passed value."""
    # print("call parser", cls, "from", tcls, "on", value, ":", type(value))
    ret = cls.parse(target, value)
    if cls.strict and not isinstance(ret, target):
        msg = f"Parser returned: {type(ret).__name__}, "
        msg += f"expected: {target.__name__} (strict=True)"
        raise RuntimeError(msg)
    return ret


def get_parser(cls):
    """Return inner Parser class, or None.

    If the inner Parser class is not a subclass of `BaseParser`,
    will raise an exception, as this is most likely an error.
    """
    if parser := cls.__dict__.get("Parser"):
        if not issubclass(parser, BaseParser):
            msg = f"{cls}: {cls.Parser.__name__} must be a subclass of {BaseParser.__name__}!"
            raise TypeError(msg)
        return parser


class NoParserDefined:
    ...


class ParserMixin:
    """Mixin class to simplify creation of custom pydantic field types.

    Can also be mixed into arbitrary classes, not just pydantic models,
    that is why it is kept separately from the top level base model we use.

    Also, we avoid using a custom metaclass for the mixin itself,
    to increase compatibility with various classes.
    """

    Parser: ClassVar[Type[BaseParser]]

    @classmethod
    def __get_validators__(cls):
        pfunc = cls.__dict__.get("__parser_func__")
        if pfunc is None:
            if parser := get_parser(cls):

                def wrapper_func(cls, value, values=None, config=None, field=None):
                    return run_parser(parser, field.type_, value)

                pfunc = wrapper_func
            else:
                pfunc = NoParserDefined
            cls.__parser_func__ = pfunc  # cache it

        if pfunc is not NoParserDefined:  # return cached parser function
            yield pfunc

        # important: if no parser is given
        # and class is a model,
        # also return the default validate function of the model!
        if issubclass(cls, BaseModel):
            yield cls.validate

    @classmethod
    def __modify_schema__(cls, schema):
        if parser := get_parser(cls):
            if schema_info := parser.schema_info:
                schema.update(**schema_info)


__all__ = ["BaseParser", "ParserMixin"]



"""Ad-hoc standalone dashboard/widget server for use within Jupyter notebooks.

It runs everything needed to see a dashboard or widget in threads.

This is mostly intended for convenient local use (e.g. by a researcher),
or could be adapted for a containerized (in the Docker-sense) environment, e.g.
where the user has metador libraries available and can inspect containers.

**Do not use this to deploy a widget server backing the widgets on a website.**
"""
import logging
import socket
from threading import Thread
from typing import List, Optional

import panel as pn
from flask import Flask

from ...container.provider import SimpleContainerProvider
from ..server import WidgetServer


def get_free_port():
    # get a free port and use it (no way to retrieve it, when letting flask choose)
    sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
    sock.bind((host, 0))
    port = sock.getsockname()[1]
    sock.close()
    return port


def silence_flask():
    """Disable HTTP request log (for use inside jupyter)."""
    log = logging.getLogger("werkzeug")
    log.setLevel(logging.ERROR)


# ----
# Use this Python module as singleton instance for server.
# As this is only used ad-hoc e.g. by a researcher playing in a notebook,
# this should be not a problem ("serious" servers are implemented elsewhere!).

DEFAULT_PANEL_EXTS = ["ace"]

host: str = "127.0.0.1"
port: int = -1

_known_containers: SimpleContainerProvider[str] = SimpleContainerProvider[str]()
_widget_server: WidgetServer = WidgetServer(_known_containers)


def widget_server() -> WidgetServer:
    return _widget_server


def container_provider() -> SimpleContainerProvider[str]:
    return _known_containers


def running() -> bool:
    return port > 0


def run(*, debug: bool = False, pn_exts: Optional[List[str]] = None):
    """Run ad-hoc standalone server to use widgets and dashboards in a Jupyter notebook."""
    global _widget_server, _known_containers, port

    if not debug:
        silence_flask()
    pn.extension(*(pn_exts or DEFAULT_PANEL_EXTS))  # required for panel within jupyter

    port = get_free_port()
    flask_base = f"http://{host}:{port}"

    # prepare bokeh server
    bokeh_port = get_free_port()
    bokeh_base = f"http://{host}:{bokeh_port}"

    def run_bokeh():
        _widget_server.run(
            host=host, port=bokeh_port, allowed_websocket_origin=[f"{host}:{port}"]
        )

    # prepare flask server
    flask_app = Flask(__name__)
    _widget_server.bokeh_endpoint = bokeh_base
    _widget_server.flask_endpoint = flask_base
    flask_bokeh = _widget_server.get_flask_blueprint("widget-api", __name__)
    flask_app.register_blueprint(flask_bokeh)

    def run_flask():
        flask_app.run(host=host, port=port)

    # launch
    t_flask = Thread(target=run_flask)
    t_bokeh = Thread(target=run_bokeh)
    t_flask.start()
    t_bokeh.start()
